
/*
*
* iscsi initiator driver
*
* Author: Kalman Meth, IBM Haifa Research Laboratory <meth@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* config.c
*
* iscsi driver configuration routines.
*
* %I% %W% %D% %T%
*/

/*
* Change History
*
*    Date     Version Defect Initials Comments
* 2001/05/26  1.11    2483   KZM      Too many disks reported if have duplicate target entry in /etc/iscsiclient.conf
*/


#ifdef __KERNEL__

#include <asm/uaccess.h>
#include "iscsi_interfaces.h"
#include "iscsi_client.h"
#include "iscsi_core_query.h"
char *clientConfigFile = ISCSI_CONFIG_FILE;

#define PRINT_PREFIX	printk("iscsi config, line %d:, pos %d, ", j, ptr-line)
#define PRINT_PREFIX2	printk("iscsi config, line %d: ", line)
#define PARSE_ERROR	printk("parse error \n")
#define ILLEGAL_CHARACTER_IN_ADDRESS	printk("illegal character in address 0%o \n", *ptr)
#define ERROR_IN_ADDRESS	printk("error in address \n")
#define ERROR_IN_PORT_NUM	printk("error in port number \n")

#else /* __KERNEL__ */

#include <asm/types.h>
#include <stdio.h>
#include "iscsi_check_config.h"

char *clientConfigFile;

#define PRINT_PREFIX	printf("line %d, pos %d, ", j, ptr-line);
#define PRINT_PREFIX2	printf("line %d: ", line)
#define PARSE_ERROR	printf("parse error \n");
#define ILLEGAL_CHARACTER_IN_ADDRESS	printf("illegal character in address 0%o \n", *ptr)
#define ERROR_IN_ADDRESS	printf("error in address \n")
#define ERROR_IN_PORT_NUM	printf("error in port number \n")

#endif /* __KERNEL__ */

#include "iscsi_local_trace.h"

#define MAX_LINE_SIZE 512


#ifdef CHECK_LOGIN_PARAMS

/*
* NAME: checkLoginParams
*
* PURPOSE: validate the login parameters
*
* PARAMETERS:
*	lineNum - line number in configuration file
*	loginParams - the login parameters
*	loginParamsLen - length (number of charachaters) of the login parameters
*
* RETURNS:
*	0 - valid
*	-1 - error
*
* CALLED BY: readConfigFile()
*
* NOTES:
* For current release, we allow only the "Initiator:" key.
* We only allow the alphanumeric characters, plus the characters
* dot ('.'), dash ('-'), and underscore ('_').
* We allow up to 64 characters for an initiator name, possibly followed
* by a comma (',') followed bu up to 16 characters of a 'password.'
* According to the iSCSI specification, we need have checked for any
* keys or for their contents. The current restriction is to make an
* administrator's life a little easier for our initial product.
* This code is shared kernel space configuration and user space parser.
*
* FUTURE ENHANCEMENTS:
* Remove restrictions on Initiator names and key values.
*
*/

#define LEGAL_CHAR(c) \
		((c >= 'a') && (c <= 'z')) || \
		((c >= 'A') && (c <= 'Z')) || \
		((c >= '0') && (c <= '9')) || \
		(c == '.') || \
		(c == '-') || \
		(c == '_')
#define MAX_INIT_ID_LEN 64
#define MAX_PASSWORD_LEN 16
/* Note: The length of "Initiator:" is 10 (w/o final '\0') */
#define INITIATOR_KEY_LEN 10

s32 checkLoginParams(u32 line, uchar *loginParams, u32 loginParamsLen)
{
	s32 i, j;
	uchar c = 0;

	if (!loginParamsLen) {
		PRINT_PREFIX2;
#ifdef __KERNEL__
		printk("missing Login Parameters\n");
#else /* __KERNEL__ */
		printf("missing Login Parameters\n");
#endif /* __KERNEL__ */
		return -1;
	}

	if ((loginParamsLen < INITIATOR_KEY_LEN) || (memcmp(loginParams, "Initiator:", INITIATOR_KEY_LEN))) {
		PRINT_PREFIX2;
#ifdef __KERNEL__
		printk("missing 'Initiator:' keyword\n");
#else /* __KERNEL__ */
		printf("missing 'Initiator:' keyword\n");
#endif /* __KERNEL__ */
		return -1;
	}
	/* check that each character is valid */
	for (i = INITIATOR_KEY_LEN; (i < loginParamsLen) && (i < (MAX_INIT_ID_LEN + INITIATOR_KEY_LEN)); i++) {
		c = loginParams[i];
		if (',' == c) {
			/* verify that we have a non-empty ID before the comma */
			if (INITIATOR_KEY_LEN == i) {
				PRINT_PREFIX2;
#ifdef __KERNEL__
				printk("missing Initiator identifier\n");
#else /* __KERNEL__ */
				printf("missing Initiator identifier\n");
#endif /* __KERNEL__ */
				return -1;
			}
			break;
		}
		if (!(LEGAL_CHAR(c))) {
			PRINT_PREFIX2;
#ifdef __KERNEL__
			printk("illegal character in identifier\n");
#else /* __KERNEL__ */
			printf("illegal character in identifier\n");
#endif /* __KERNEL__ */
			return -1;
		}
	}
	/* verify that current character is a comma */
	if (',' != c) {
		PRINT_PREFIX2;
#ifdef __KERNEL__
		if (i < (MAX_INIT_ID_LEN + INITIATOR_KEY_LEN)) {
			printk("missing ',' after identifier\n");
		}
		else if (LEGAL_CHAR(c)) {
			/* identifier is too long */
			printk("identifier is too long\n");
		}
		else {
			printk("missing ',' after identifier\n");
		}
#else /* __KERNEL__ */
		if (i < (MAX_INIT_ID_LEN + INITIATOR_KEY_LEN)) {
			printf("missing ',' after identifier\n");
		}
		else if (LEGAL_CHAR(c)) {
			/* identifier is too long */
			printf("identifier is too long\n");
		}
		else {
			printf("missing ',' after identifier\n");
		}
#endif /* __KERNEL__ */
		return -1;
	}
	/* verify that 'password' is valid */
	for (j = i+1; (j < loginParamsLen) && (j <= (i + MAX_PASSWORD_LEN)); j++) {
		c = loginParams[j];
		if (!(LEGAL_CHAR(c))) {
			PRINT_PREFIX2;
#ifdef __KERNEL__
			printk("illegal character in 'password'\n");
#else /* __KERNEL__ */
			printf("illegal character in 'password'\n");
#endif /* __KERNEL__ */
			return -1;
		}
	}
	if (j != loginParamsLen) {
		PRINT_PREFIX2;
		c = loginParams[j];
#ifdef __KERNEL__
		if (!(LEGAL_CHAR(c))) {
			printk("illegal characters after 'password'\n");
		}
		else {
			printk("'password' is too long\n");
		}
#else /* __KERNEL__ */
		if (!(LEGAL_CHAR(c))) {
			printf("illegal characters after 'password'\n");
		}
		else {
			printf("'password' is too long\n");
		}
#endif /* __KERNEL__ */
		return -1;
	}
	return 0;
}

#endif /* CHECK_LOGIN_PARAMS */


/*
* NAME: readConfigFile
*
* PURPOSE: read config file and perform calls to create sessions
*
* PARAMETERS: none
*
* RETURNS:
*	number of targets identified in configuration file
*	-1 for error
*
* CALLED BY: iscsi_detect()
*
* NOTES:
* read client (initiator) configuration file.
* make connections, if possible.
* return number of targets defined.
*/
s32 readConfigFile(void)
{
	uchar line[MAX_LINE_SIZE];
	u32 nConnections;
	uchar *loginParams = NULL;
	u32 loginParamsLen;
	u16 portNum;
	s32 rc;
	int i, j;
	uchar c;
	uchar n1, n2, n3, n4;
	u32 addr;
	uchar addr1[4];
	uchar *ptr;
	uchar *ptr2;
#ifdef __KERNEL__
	struct file *fptr;
	mm_segment_t oldfs;
	querySession_t qsData;
#else /* __KERNEL__ */
	FILE *fptr;
#endif /* __KERNEL__ */
	s32 nTargets = 0;

	TRACE_ENTER;

	/* open the client configuration file */

#ifdef __KERNEL__
	fptr = filp_open(clientConfigFile, O_RDONLY, 0);
	if (IS_ERR(fptr)) {
		return 0;
	}
#else /* __KERNEL__ */
	fptr = fopen(clientConfigFile, "r");
	if (NULL == fptr) {
		fprintf(stderr, "failed to open file %s\n", clientConfigFile);
		return 0;
	}
#endif /* __KERNEL__ */

	/* j is the line number in configuration file */
	for (j = 1; 1; j++) {
		/* PR1("line number = %d \n", j); */
		/*
		Read in a line from the configuration file.
		We read one character at a time until we hit an end-of-line.
		*/
		/* i is the location within the line */
		for (i = 0; i < MAX_LINE_SIZE; i++) {
#ifdef __KERNEL__
			oldfs = get_fs();
			set_fs(get_ds()); /* read buffer is in kernel space */
			rc = fptr->f_op->read(fptr, &c, 1, &fptr->f_pos);
			set_fs(oldfs);
			if (rc < 0) {
				filp_close(fptr, current->files);
				return rc;
			}
#else /* __KERNEL__ */
			rc = fread(&c, 1, 1, fptr);
#endif /* __KERNEL__ */
			if (0 == rc) {
				/* EOF */
				if (!i) {
					/* at beginning of a line; nothing more to do */
#ifdef __KERNEL__
					filp_close(fptr, current->files);
#else /* __KERNEL__ */
					fclose(fptr);
#endif /* __KERNEL__ */
					TRACE_EXIT;
					PRINT(nTargets);
					return nTargets;
				}
				/* finish processing existing line */
				c = '\n';
			}
			line[i] = c;
			if ('\n' == c) {
				/* reached end of line */
				/* PR2("reached end of line %d, length = %d \n", j, i); */
				line[i] = '\0';
				break;
			}
		}
		/* PR1("line number = %d\n", j); */
		/* PR1("line length = %d\n", i); */
		if (!i) {
			/* empty line */
			continue;
		}

		/* process the line */

		if ('#' == line[0]) {
			/* this is a comment line; skip it */
			continue;
		}

		/* read target info */
		ptr = line;
		/* eat up initial white space */
		for (; (' ' == *ptr) || ('\t' == *ptr); ptr++) { }

		/* verify that the prefix "target=" is present */
		if (!(('t' == ptr[0]) && ('a' == ptr[1]) && ('r' == ptr[2]) && ('g' == ptr[3]) && ('e' == ptr[4]) && ('t' == ptr[5]))) {
			/* parsing error; skip over line */
			PRMSG("parse error");
			PRINT_PREFIX;
			PARSE_ERROR;
			continue;
		}
		ptr += 6;

		/* eat up white space */
		for (; (' ' == *ptr) || ('\t' == *ptr); ptr++) { }

		/* eat up the '=' */
		if ('=' != *ptr) {
			/* parsing error; skip over line */
			PRMSG("parse error");
			PRINT_PREFIX;
			PARSE_ERROR;
			continue;
		}
		ptr++;

		/* eat up white space */
		for (; (' ' == *ptr) || ('\t' == *ptr); ptr++) { }

		n1 = 0;
		ptr2 = ptr;
		while ((*ptr >= '0') && (*ptr <= '9')) {
			n1 = (n1 * 10) + (*ptr - '0');
			ptr++;
		}
		if (ptr2 == ptr) {
			PRINT_PREFIX;
			ERROR_IN_ADDRESS;
			continue;
		}
		if (*ptr != '.') {
			PRINT_PREFIX;
			ILLEGAL_CHARACTER_IN_ADDRESS;
			continue;
		}
		ptr++;
		n2 = 0;
		ptr2 = ptr;
		while ((*ptr >= '0') && (*ptr <= '9')) {
			n2 = (n2 * 10) + (*ptr - '0');
			ptr++;
		}
		if (ptr2 == ptr) {
			PRINT_PREFIX;
			ERROR_IN_ADDRESS;
			continue;
		}
		if (*ptr != '.') {
			PR1("illegal character in address %c \n", *ptr);
			PRINT_PREFIX;
			ILLEGAL_CHARACTER_IN_ADDRESS;
			continue;
		}
		ptr++;
		n3 = 0;
		ptr2 = ptr;
		while ((*ptr >= '0') && (*ptr <= '9')) {
			n3 = (n3 * 10) + (*ptr - '0');
			ptr++;
		}
		if (ptr2 == ptr) {
			PRINT_PREFIX;
			ERROR_IN_ADDRESS;
			continue;
		}
		if (*ptr != '.') {
			PR1("illegal character in address %c \n", *ptr);
			PRINT_PREFIX;
			ILLEGAL_CHARACTER_IN_ADDRESS;
			continue;
		}
		ptr++;
		n4 = 0;
		ptr2 = ptr;
		while ((*ptr >= '0') && (*ptr <= '9')) {
			n4 = (n4 * 10) + (*ptr - '0');
			ptr++;
		}
		if (ptr2 == ptr) {
			PRINT_PREFIX;
			ERROR_IN_ADDRESS;
			continue;
		}

		/* place address components into character array in network byte order */
		addr1[0] = n1;
		addr1[1] = n2;
		addr1[2] = n3;
		addr1[3] = n4;
		memcpy(&addr, addr1, 4);
		PRHEX(addr);

		/* if next character is not space, have error */
		if (' ' != *ptr) {
			PRINT_PREFIX;
			ERROR_IN_ADDRESS;
			continue;
		}

		/* eat up white space */
		for (; (' ' == *ptr) || ('\t' == *ptr); ptr++) { }

		/* extract port number */
		portNum = 0;
		ptr2 = ptr;
		while ((*ptr >= '0') && (*ptr <= '9')) {
			portNum = (portNum * 10) + (*ptr - '0');
			ptr++;
		}
		if (ptr2 == ptr) {
			PRINT_PREFIX;
			ERROR_IN_PORT_NUM;
			continue;
		}
		PRINT(portNum);

		/* if next character is not space, have error */
		if (' ' != *ptr) {
			PRINT_PREFIX;
			ERROR_IN_PORT_NUM;
			continue;
		}

		/* eat up white space */
		for (; (' ' == *ptr) || ('\t' == *ptr); ptr++) { }

		/* rest of line is optional: nConnections and parameters */
		loginParamsLen = 0;
		nConnections = 1;
		if (*ptr != '\0') {
			nConnections = 0;
			/* obtain number of connections */
			PRINT(nConnections);
			while ((*ptr >= '0') && (*ptr <= '9')) {
				nConnections = (nConnections * 10) + (*ptr - '0');
				ptr++;
			}
			PRINT(nConnections);

			/* eat up white space */
			for (; (' ' == *ptr) || ('\t' == *ptr); ptr++) { }
		}
		if (*ptr != '\0') {
			loginParams = ptr;
			/* figure out length of parameters; read until white space or EOL (which was changed to '\0') */
			for (i = 0; (' ' != *ptr) && ('\t' != *ptr) && ('\0' != *ptr); ptr++, i++) { }
			loginParamsLen = i;
		}
		PRINT(loginParamsLen);

#ifdef CHECK_LOGIN_PARAMS
		rc = checkLoginParams(j, loginParams, loginParamsLen);
		if (rc < 0) {
			continue;
		}
#endif /* CHECK_LOGIN_PARAMS */

		if (!nConnections) {
			PRINT_PREFIX;
#ifdef __KERNEL__
			printk("no connections defined \n");
#else /* __KERNEL__ */
			printf("no connections defined \n");
#endif /* __KERNEL__ */
			continue;
		}

		/* create session */
#ifdef __KERNEL__
		printk("iscsi config, line %d, nConnections = %d, target addr = %d.%d.%d.%d, portNum = %d, loginParamsLen = %d, loginParams = %s \n", j, nConnections, n1, n2, n3, n4, portNum, loginParamsLen, loginParams);
		rc = iscsiCoreCreateSession(addr, portNum, nConnections, loginParams, loginParamsLen);
		PRINT(rc);
		if (rc >= 0) {
			/*
			* It is possible that a user will try to define
			* the same session more than once.
			* We will get back the same session index.
			* Verify that the session index is not already
			* in an earlier entry of sessionIndeces[].
			*/
			s32 duplicate = 0;
			for (i = 0; i < nTargets; i++) {
				if (sessionIndeces[i] == rc) {
					/* have duplicate */
					printk("iscsi config, line %d, duplicate entry \n", j);
					duplicate = 1;
				}
			}
			if (duplicate) {
				/* go to next entry */
				continue;
			}
			/* MRN - CMVC#973 - Add Error Message */
			memset(&qsData, 0, sizeof(querySession_t));
			iscsiCoreQuerySession(rc, &qsData);
			if (qsData.numActiveConnections == 0) {
				printk("Target %d.%d.%d.%d unavailable or no virtual disks configured for this initiator.\n", n1, n2, n3, n4);
			}
			/* MRN - END CMVC#973 */
         		sessionIndeces[nTargets] = rc;
			nTargets++;
		}
		else {
			printk("iscsi config, failed to create session\n");
		}
#else /* __KERNEL__ */
		printf("line %d, nConnections = %d, target addr = %d.%d.%d.%d, portNum = %d, loginParamsLen = %d, loginParams = %s \n", j, nConnections, n1, n2, n3, n4, portNum, loginParamsLen, loginParams);
#endif /* __KERNEL__ */
	}
	TRACE_EXIT;
}
