
/*
*
* iscsi initiator driver
*
* Author: Uri Schonfeld, IBM Haifa Research Laboratory <shuri@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* iscsi_ioctl.c
*
* iscsi driver ioctl-s.
*
* %I% %W% %D% %T%
*/


/* Includes */
#ifndef __KERNEL__
#  define __KERNEL__
#endif /* __KERNEL__ */
#ifndef MODULE
#  define  MODULE
#endif /* MODULE */

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/blkdev.h>
#include <asm/uaccess.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>   /* kernel_thread */
#include <linux/fs.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <asm/page.h>
#include <asm/processor.h>
#include <scsi.h>
#include <hosts.h>
#include <sd.h>
#include "iscsi_interfaces.h"
#include "iscsi_client.h"
#include "iscsi_ioctl.h"
#include "iscsi_local_trace.h"
#include "iscsi_core_query.h"


/* Memory access abstraction definitions */
#ifdef LINUX_20
#  include <asm/segment.h>
#  define access_ok(t,a,sz)           (verify_area((t),(a),(sz)) ? 0 : 1)
#  define verify_area_20              verify_area
#  define copy_to_user(t,f,n)         (memcpy_tofs(t,f,n), 0)
#  define __copy_to_user(t,f,n)       copy_to_user((t),(f),(n))
#  define copy_to_user_ret(t,f,n,r)   copy_to_user((t),(f),(n))
#  define copy_from_user(t,f,n)       (memcpy_fromfs((t),(f),(n)), 0)
#  define __copy_from_user(t,f,n)     copy_from_user((t),(f),(n))
#  define copy_from_user_ret(t,f,n,r) copy_from_user((t),(f),(n))
#  define PUT_USER(val,add)           (put_user((val),(add)), 0)
#  define __PUT_USER(val,add)         PUT_USER((val),(add))
#  define PUT_USER_RET(val,add,ret)   PUT_USER((val),(add))
#  define GET_USER(dest,add)          ((dest)=get_user((add)), 0)
#  define __GET_USER(dest,add)        GET_USER((dest),(add))
#  define GET_USER_RET(dest,add,ret)  GET_USER((dest),(add))
#else
#  include <asm/uaccess.h>
#  include <asm/io.h>
#  define verify_area_20(t,a,sz) (0) /* == success */
#  define PUT_USER put_user
#  define __PUT_USER __put_user
#  define PUT_USER_RET put_user_ret
#  define GET_USER get_user
#  define __GET_USER __get_user
#  define GET_USER_RET get_user_ret
#endif

/* Function Headers */
int importTargetEntry(Target_Entry *, void *);
int iscsi_ioctl_addTarget(Target_Entry *);
int iscsi_ioctl_deleteTarget(int );
int iscsi_ioctl_resetBus(int);

/*********************************************************************
 * NAME: iscsi_ioctl
 *
 * PURPOSE:    
 *          
 *          
 *
 * PARAMETERS:
 *	Cmnd    - integer corresponding to the ioctl to be performed
 *      arg     - pointer to a user space Target_Entry structure
 *
 * RETURNS:
 *      0         - for success
 *      Otherwise - An appropriate negative error code
 *
 * CALLED BY: system
 *
 * NOTES:
 * 
 *********************************************************************/
int iscsi_ioctl(struct inode *inode, struct file *file, unsigned int Cmnd, 
		unsigned long arg){ 

  Target_Entry target;
  int rc;

  TRACE_ENTER;
  PRHEX(Cmnd);

  /* Init rc to Zero */
  /* Returning a non zero value, will mark the device used, 
   * preventing unloading of module */
  rc = 0;
  memset(&target, 0, sizeof(Target_Entry));

  /* Handle Command */
  switch(Cmnd){
  case ISCSI_IOCTL_ADD_TARGET:
     PRMSG("DBUG: Entered ISCSI_IOCTL_ADD_TARGET");
    /* Import Target_Entry from user space */
    rc = importTargetEntry(&target, (void *)arg);
    if (rc != 0){
      PRMSG("DBUG: Failed to import user-space arguments");
      if (NULL != target.loginParams) {
        iscsiOSfree(target.loginParams);
      }
      return rc;
    }
    
    /* And perform add */
    rc = iscsi_ioctl_addTarget(&target);
    if (rc != 0){
      PRMSG("DBUG: Failed iscsi_ioctl_addTarget");
    }
    if (NULL != target.loginParams) {
      iscsiOSfree(target.loginParams);
    }
    break;
  case ISCSI_IOCTL_DEL_TARGET:
    PRMSG("DBUG: ISCSI_IOCTL_DEL_TARGET");
    /* Delete the target */
    rc = iscsi_ioctl_deleteTarget((int)arg);
    if (rc != 0){
      PRMSG("DBUG: Failed iscsi_ioctl_deleteTarget");
    }
    break;

  case ISCSI_IOCTL_RESET_BUS:
    PRMSG("DBUG: ISCSI_IOCTL_RESET_BUS");
    rc = iscsi_ioctl_resetBus((int) arg);
    if (rc != 0){
      PRMSG("DBUG: Failed iscsi_ioctl_resetBus");
    }
    break;
  case ISCSI_IOCTL_RESET_DRIVER:
    PRMSG("DBUG: ISCSI_IOCTL_RESET_DRIVER");
    iscsiCoreResetSessions();
    rc = 0;
    break;
  case ISCSI_IOCTL_QUERY_SESSION:
    PRMSG("DBUG: ISCSI_IOCTL_QUERY_SESSION");
    {
    iscsiQueryIo_t iqio;
    querySession_t qs;

    copy_from_user(&iqio, (void*) arg, sizeof(iscsiQueryIo_t));
    rc = iscsiCoreQuerySession(iqio.sessionNumber, &qs);
    if (!rc) {
      copy_to_user(iqio.buf, &qs, sizeof(querySession_t));
    }
    break;
    }
  case ISCSI_IOCTL_QUERY_CONNECTION:
    PRMSG("DBUG: ISCSI_IOCTL_QUERY_CONNECTION");
    {
    iscsiQueryIo_t iqio;
    queryConnection_t qc;

    copy_from_user(&iqio, (void*) arg, sizeof(iscsiQueryIo_t));
    copy_from_user(&qc, (void*) iqio.buf, sizeof(queryConnection_t));
    rc = iscsiCoreQueryConnection(iqio.sessionNumber, &qc);
    if (!rc) {
      copy_to_user(iqio.buf, &qc, sizeof(queryConnection_t));
    }
    break;
    }
  case ISCSI_IOCTL_SESSION_COUNTERS:
    PRMSG("DBUG: ISCSI_IOCTL_SESSION_COUNTERS");
    {
    iscsiQueryIo_t iqio;
    sessionStats_t stats;

    copy_from_user(&iqio, (void*) arg, sizeof(iscsiQueryIo_t));
    rc = iscsiCoreGetSessionStats(iqio.sessionNumber, &stats);
    if (!rc) {
      copy_to_user(iqio.buf, &stats, sizeof(sessionStats_t));
    }
    break;
    }
  default:
    printk("illegal iscsi ioctl command = 0x%x \n", Cmnd);
    rc = -1;
    break;
  }/* EOF switch(Cmnd) */


  if (rc != 0){
    PRMSG("Could'nt perform ioctl successfully");
  }

  TRACE_EXIT;
  return rc;
}/* EOF iscsi_ioctl */
    

/*********************************************************************
 * NAME: iscsi_ioctl_addTarget
 *
 * PURPOSE:    Peform all that is needed to add a new target: both 
 *          at mid-level and low-lvel
 *          
 *          
 *
 * PARAMETERS:
 *	target - pointer to an allocated Target_Entry structure
 *
 * RETURNS:
 *      0         - for success
 *      Otherwise - An appropriate negative error code
 *
 * CALLED BY: iscsi_ioctl
 *
 * NOTES:
 * 
 *********************************************************************/

int iscsi_ioctl_addTarget(Target_Entry *target){
  
  int rc;
#ifdef CONFIG_ISCSI_KERNEL
  s32 sessionNum;
  int i;
#endif //CONFIG_ISCSI_KERNEL

  TRACE_ENTER;

  rc = 0;

#ifdef CONFIG_ISCSI_KERNEL

  /* add target at low level driver */
  sessionNum = iscsiCoreCreateSession(target->addr, target->portNum, target->nConnections, target->loginParams, target->loginParamsLen);
  if (sessionNum == -1){
    rc = -1;
    return rc;
  }

  /* Find an empty sessionIndeces entry */
  for (i=0; i<ISCSI_MAX_TARGETS; i++){
    if (sessionIndeces[i] == -1){
      break;
    }
  }
  if (i == ISCSI_MAX_TARGETS){
    printk("No more iSCSI sessions available");
    return -1;
  }

  /* Mark it as active */
  sessionIndeces[i] = sessionNum;

  printk("Created Session %d\n", sessionNum);

  /* Rescan Target at Mid-Level Driver the appropriate target/channel */
  rc = add_multiple_devices(iSCSI_Host, i, -1, -1); 
#else
  printk("You can not dynamically add a target. You have to compile the linux kernel with scsi rescan support, and the iscsiclient with CONFIG_ISCSI_KERNEL\n");
#endif //CONFIG_ISCSI_KERNEL

  TRACE_EXIT;
  return rc;

}/* EOF iscsi_ioctl_addTarget */

/*********************************************************************
 * NAME: iscsi_ioctl_deleteTarget
 *
 * PURPOSE:    Peform all that is needed to delete a target: both 
 *          at mid-level and low-lvel
 *          
 *          
 *
 * PARAMETERS:
 *	target - integer corresponding to the SessionIndeces index
 *               and the channel associated with the target
 *
 * RETURNS:
 *      0         - for success
 *      Otherwise - An appropriate negative error code
 *
 * CALLED BY: iscsi_ioctl
 *
 * NOTES:
 * 
 *********************************************************************/
int iscsi_ioctl_deleteTarget(int target){
  
  int rc;
 
  TRACE_ENTER;

  /* So far so good :> */
  rc = 0;

#ifdef CONFIG_ISCSI_KERNEL
  /* Delete from mid-level/high-level */
  rc = remove_multiple_devices(iSCSI_Host, target, -1, -1); 
  /* If Unsuccessful Stop here */
  if (rc != 0){
    return rc;
  }

  /* Successful, delete from low-level as well */
  /* Note: target is assumed to be the channel number and session handle */
  rc =  iscsiCoreDeleteSession(target);

  if (rc == 0){
	  sessionIndeces[target] = -1;
  }
#else
  printk("You can not dynamically delete a target. You have to compile the linux kernel with scsi rescan support, and the iscsiclient with CONFIG_ISCSI_KERNEL\n");
#endif //CONFIG_ISCSI_KERNEL
  
  TRACE_EXIT;
  return rc;

}/* EOF iscsi_ioctl_deleteTarget */





/*********************************************************************
 * NAME: iscsi_ioctl_resetBus
 *
 * PURPOSE:    Peform complete reset of iscsi bus
 *          
 *          
 *
 * PARAMETERS:NONE
 *
 * RETURNS:
 *      0         - for success
 *      Otherwise - An appropriate negative error code
 *
 * CALLED BY: iscsi_ioctl
 *
 * NOTES:
 * 
 *********************************************************************/

int iscsi_ioctl_resetBus(int target){
  
  int rc;

  /* All is well, so far */
  rc = 0;

  /* Note: target is assumed to be the channel number and session handle */
  rc =  iscsiCoreResetBus(target, NULL, ISCSI_WAIT_FOR_COMPLETION);
  if (rc == ISCSI_RESET_SUCCESS){
    rc =0;
  } else {
    rc = -1;
  }

  return (rc);

}/* EOF iscsi_ioctl_resetBus() */



/*********************************************************************
 * NAME: importTargetEntry
 *
 * PURPOSE:    Import Target Entry from a pointer to a Target_Entry 
 *          user space structure.
 *          
 *
 * PARAMETERS:
 *	target - pointer to an allocated Target_Entry structure
 *      arg    - pointer to a user space Target_Entry structure
 *
 * RETURNS:
 *      0         - for success
 *      Otherwise - An appropriate negative error code
 * CALLED BY: iscsi_ioctl
 *
 * NOTES:
 * 
 *********************************************************************/
int importTargetEntry(Target_Entry *target, void *arg){

  Target_Entry *userTarget;
  void *userLoginParams;

  TRACE_ENTER;
  userTarget = (Target_Entry *)arg;

  /* Get Struct from user */
  copy_from_user(target, arg, sizeof(Target_Entry));

  /* Allocate mem for loginParams */
  userLoginParams = target->loginParams;
  target->loginParams = iscsiOSmalloc(sizeof(uchar)*(target->loginParamsLen));
  if (target->loginParams == NULL){
    return -ENOMEM;
  }

  /* Copy loginParams */
  copy_from_user(target->loginParams, userLoginParams, target->loginParamsLen);
  PRHEX(target->addr);
  PRINT(target->portNum);
  PRINT(target->nConnections);
  PRINT(target->loginParamsLen);
  PRBUF(target->loginParams, target->loginParamsLen);
  
  TRACE_EXIT;
  return 0;

}/* EOF importTargetEntry */

