
/*
*
* iscsi initiator driver
*
* Author: Kalman Meth, IBM Haifa Research Laboratory <meth@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* messages.c
*
* Routines to manipulate iSCSI message headers and pending commands.
*
* %I% %W% %D% %T%
*/

#include "iscsi_local_types.h"
#include "iscsi_interfaces.h"
#include "iscsi_core_messages.h"
#include "iscsi_local_trace.h"
#include "iscsi_core_prot.h"


/* ************************************************ */

/*
* In current implementation, pending commands are stored
* in a doubly linked list.
* In the future, we may want to change the pending command table
* into a hash table instead of a doubly linked list.
* In order to be able to ensure that returned TaskIds are valid,
* we should verify that the TaskId is in fact valid in the
* pending command table.
*/

/*
* NAME: initPendingCmdTables
*
* PURPOSE: initialize table of pending commands and Task Management for specified session
*
* PARAMETERS:
*	si - pointer to session information
*
* RETURNS: none
*
* CALLED BY: iscsiCoreCreateSession()
*
* NOTES:
*/
void initPendingCmdTable(sessionInfo_t *si)
{
	pendCmdTable_t	*pct = &si->pendCmdTable;
	pendTaskManTable_t	*ptmt = &si->pendTaskManTable;

	pct->base.forw = &pct->base;
	pct->base.back = &pct->base;
	pct->head = &pct->base;
	iscsiOSlockInit(&pct->lock);

	ptmt->base.forw = &ptmt->base;
	ptmt->base.back = &ptmt->base;
	ptmt->head = &ptmt->base;
	iscsiOSlockInit(&ptmt->lock);
}



/*
* NAME: enterCmdIntoTable
*
* PURPOSE: enter specified pending command into specified session's command table
*
* PARAMETERS:
*	si - pointer to session information
*	cmd - poiner to pending command information
*
* RETURNS: none
*
* CALLED BY: sendCmd()
*
* NOTES:
* Place item at end of doubly linked list.
* Once an item is placed in the table, it is only removed
* when either a response has arrived for the item, or the
* connection related to the item has been broken.
* If the command is subsequently aborted, it is not removed
* from the table until we get confirmation from the target
* that it has been aborted.
*/
void enterCmdIntoTable(sessionInfo_t *si, pendingCmd_t *cmd)
{
	iscsiIrql_t irql;
	pendCmdTable_t *pcl = &si->pendCmdTable;

	TRACE_ENTER;
	iscsiOSlock(&pcl->lock, &irql);
	cmd->forw = pcl->head;
	cmd->back = pcl->head->back;
	pcl->head->back->forw = cmd;
	pcl->head->back = cmd;
	iscsiOSunlock(&pcl->lock, irql);
	TRACE_EXIT;
	return;
}



/*
* NAME: removeCmdFromTable
*
* PURPOSE: remove pending command structure from session's table
*
* PARAMETERS:
*	si - pointer to session structure
*	cmd - pointer to pending command structure to be removed
*
* RETURNS:
*	pointer to structure removed, if found in table
*	NULL, if not found in table.
*
* CALLED BY: handleScsiResponse(), handleDataRead(), etc
*
* NOTES:
* Remove item from table.
* Verify that the command is in fact in the table.
*/
pendingCmd_t *removeCmdFromTable(sessionInfo_t *si, pendingCmd_t *cmd)
{
	iscsiIrql_t irql;
	pendingCmd_t *ptr;
	pendCmdTable_t *pct;

	TRACE_ENTER;
	pct = &si->pendCmdTable;
	iscsiOSlock(&pct->lock, &irql);
	for (ptr = pct->head->forw; ptr != pct->head; ptr = ptr->forw) {
		PRHEX(ptr);
		if (ptr == cmd) {
			cmd->back->forw = cmd->forw;
			cmd->forw->back = cmd->back;
			cmd->forw = NULL;
			cmd->back = NULL;
		}
		iscsiOSunlock(&pct->lock, irql);
		TRACE_EXIT;
		return ptr;
	}
	iscsiOSunlock(&pct->lock, irql);
	TRACE_EXIT;
	return NULL;
}



/*
* NAME: removeCmdsResetLun
*
* PURPOSE: remove pending commands structure for specified reset device
*
* PARAMETERS:
*	si - pointer to session structure
*	lun - device identifier
*
* RETURNS: none
*
* CALLED BY: handleTaskManResponse()
*
* NOTES:
* Scan entire table of pending commands and remove those items that
* reference the specified LUN. Complete the identified commands with
* an indication that a reset occured.
* xxx Should we also scan the task management table and remove
* references to abort commands related to this LUN?
*/
void removeCmdsResetLun(sessionInfo_t *si, iscsiLUN_t lun)
{
	pendingCmd_t	*cmd;
	pendCmdTable_t	*pct;
	iscsiIrql_t	irql;

	pct = &si->pendCmdTable;
restart:
	iscsiOSlock(&pct->lock, &irql);
	for (cmd = pct->head->forw; cmd != pct->head; ) {
		if (SAME_LUN(lun, cmd->lun)) {
			cmd->back->forw = cmd->forw;
			cmd->forw->back = cmd->back;

			/* don't hold lock when calling OS dependent function */
			iscsiOSunlock(&pct->lock, irql);
			iscsiOScompleteCommandReset(cmd->scb);
			ISCSI_INCREMENT_COUNTER(si, nCmdsReset);
			Qput(&freeCmdQ, cmd, pendingCmd_t);
			goto restart;
		}
		else {
			cmd = cmd->forw;
		}
	}
	iscsiOSunlock(&pct->lock, irql);
}


/*
* NAME: removeCmdsBadConnection
*
* PURPOSE: remove pending commands structure for specified connection
*
* PARAMETERS:
*	si - pointer to session structure
*	ci - pointer to connection information
*
* RETURNS: none
*
* CALLED BY: readConnectionHandler()
*
* NOTES:
* Scan entire table of pending commands and remove those items that
* reference the specified connection. Complete the identified commands
* with an indication that we have a connection problem.
*
* FUTURE ENHANCEMENTS:
* In a future implementation, we may try to resend these commands over
* another connection.
*/
void removeCmdsBadConnection(sessionInfo_t *si, connInfo_t *ci)
{
	pendingCmd_t	*cmd;
	pendCmdTable_t	*pct;
	iscsiIrql_t	irql;

	pct = &si->pendCmdTable;
restart:
	iscsiOSlock(&pct->lock, &irql);
	for (cmd = pct->head->forw; cmd != pct->head; ) {
		if (cmd->connInfo == ci) {
			cmd->back->forw = cmd->forw;
			cmd->forw->back = cmd->back;

			/* don't hold lock when calling OS dependent function */
			iscsiOSunlock(&pct->lock, irql);
			iscsiOScompleteCommandReset(cmd->scb);

			/* complete command with error indication */
			/* LOGICAL UNIT COMMUNICATION ERROR (0x08, 0x00) */
			// SET_SENSE(senseData, ISCSI_SCSISENSE_ABORTED_COMMAND, 0x08, 0x00);
			// iscsiOSputSense(cmd->scb, senseData, sizeof(senseData_t));
			// iscsiOScompleteCommand(cmd->scb, ISCSI_SCSISTAT_CHECK_CONDITION);
			Qput(&freeCmdQ, cmd, pendingCmd_t);
			goto restart;
		}
		else {
			cmd = cmd->forw;
		}
	}
	iscsiOSunlock(&pct->lock, irql);
}


/*
* NAME: removeTaskManBadConnection
*
* PURPOSE: remove Task Management structures for specified connection
*
* PARAMETERS:
*	si - pointer to session structure
*	ci - pointer to connection information
*
* RETURNS: none
*
* CALLED BY: readConnectionHandler()
*
* NOTES:
* Scan table of pending Task Management operations and remove those
* items that reference the specified connection. Complete the identified
* operations with an error indication.
*/
void removeTaskManBadConnection(sessionInfo_t *si, connInfo_t *ci)
{
	taskManageInfo_t	*tm;
	pendTaskManTable_t	*ptmt;
	iscsiIrql_t	irql;

	ptmt = &si->pendTaskManTable;
restart:
	iscsiOSlock(&ptmt->lock, &irql);
	for (tm = ptmt->head->forw; tm != ptmt->head; ) {
		if (tm->connInfo == ci) {
			tm->back->forw = tm->forw;
			tm->forw->back = tm->back;
			iscsiOSunlock(&ptmt->lock, irql);
			/* complete operation with error indication */
			if (ABORT_TASK == tm->operation) {
				tm->result = ISCSI_ABORT_ERROR;
			}
			else {
				/* xxx this may change */
				tm->result = ISCSI_RESET_ERROR;
			}
			finishTaskManOperation(si, tm);
			goto restart;
		}
		else {
			tm = tm->forw;
		}
	}
	iscsiOSunlock(&ptmt->lock, irql);
}


/*
* NAME: findCmdInTable
*
* PURPOSE: find pending command structure of specified SCSI Command Block
*
* PARAMETERS:
*	si - pointer to session structure
*	scb - pointer to desired SCSI Conrol Block
*
* RETURNS: pointer to corresponding pending command structure; NULL if not found
*
* CALLED BY: iscsiCoreAbortCommand()
*
* NOTES:
*/
pendingCmd_t *findCmdInTable(sessionInfo_t *si, SCB_t *scb)
{
	pendingCmd_t	*cmd;
	pendCmdTable_t	*pct;
	iscsiIrql_t	irql;

	TRACE_ENTER;
	pct = &si->pendCmdTable;;
	iscsiOSlock(&pct->lock, &irql);
	for (cmd = pct->head->forw; cmd != pct->head; cmd = cmd->forw) {
		if (cmd->scb == scb) {
			/* 
			Note: There might be a race condition between a response coming for the
			command and sending out an abort message using a canceled (completed)
			cmd. The same command sructure may then be used for a different command,
			and the abort operation may be thought to be intended for that new command.
			This is very unlikely. In any case, the target should not find the command
			(since it already completed),
			so the abort operation will fail, causing no harm.
			*/
			break;
		}
	}
	iscsiOSunlock(&pct->lock, irql);
	if (cmd == pct->head) {
		/* command was not found */
		cmd = NULL;
	}
	PRHEX(cmd);
	TRACE_EXIT;
	return cmd;
}


/*
* NAME: cleanupPendingOpsTables
*
* PURPOSE: clean up commands and Task Management operations remaining in session's tables
*
* PARAMETERS:
*	si - pointer to session information structure
*
* RETURNS: none
*
* CALLED BY: freeSessionResources()
*
* NOTES:
* Complete each pending command with an error indication.
* This function is called only after the session's threads have exited.
* There is no contention for lock of pending commands table.
*/
void cleanupPendingOpsTables(sessionInfo_t *si)
{
	pendCmdTable_t	*pct = &si->pendCmdTable;
	pendTaskManTable_t	*ptmt = &si->pendTaskManTable;
	pendingCmd_t	*cmd;
	taskManageInfo_t	*tm;
	iscsiIrql_t irql;

	TRACE_ENTER;

	/* empty out the pending command table */
restart:
	iscsiOSlock(&pct->lock, &irql);
	for (cmd = pct->head->forw; cmd != pct->head; ) {
		PRHEX(cmd);
		PRHEX(cmd->scb);
		/* remove item from table */
		cmd->back->forw = cmd->forw;
		cmd->forw->back = cmd->back;

		/* complete the command with an error indication */
		/* don't hold lock when calling OS dependent function */
		iscsiOSunlock(&pct->lock, irql);
		iscsiOScompleteCommandReset(cmd->scb);
		ISCSI_INCREMENT_COUNTER(si, nCmdsReset);

		/* return resources of message to pool */
		Qput(&freeMsgHdrQ, cmd->header, iscsiHeader_t);
		Qput(&freeCmdQ, cmd, pendingCmd_t);
		goto restart;
	}
	iscsiOSunlock(&pct->lock, irql);

	iscsiOSlockDestroy(&pct->lock);

	/* empty out the pending Task Mangement table */
restart2:
	iscsiOSlock(&ptmt->lock, &irql);
	for (tm = ptmt->head->forw; tm != ptmt->head; ) {
		PRHEX(tm);
		/* remove item from table */
		tm->back->forw = tm->forw;
		tm->forw->back = tm->back;

		/* complete the operation with an error indication */
		/* don't hold lock when calling OS dependent function */
		iscsiOSunlock(&ptmt->lock, irql);
		switch (tm->operation) {
			case ABORT_TASK:
				tm->result = ISCSI_ABORT_ERROR;
				break;
			default:
				/* xxx update this when we support operations besides RESET */ 
				tm->result = ISCSI_RESET_ERROR;
				break;
		}
		finishTaskManOperation(si, tm);
		goto restart2;
	}
	iscsiOSunlock(&ptmt->lock, irql);
	iscsiOSlockDestroy(&ptmt->lock);

	TRACE_EXIT;
}


/* ************************************************ */


/*
* NAME: getMessageHeader
*
* PURPOSE: return pointer to a message header structure
*
* PARAMETERS: none
*
* RETURNS: pointer to a message header structure
*
* CALLED BY: sendCmd(), sendRttData(), etc
*
* NOTES:
*/
iscsiHeader_t *getMessageHeader(void) {
	iscsiHeader_t *message;

	TRACE_ENTER;
	Qget(&freeMsgHdrQ, message, iscsiHeader_t);
	if (NULL == message) {
		message = iscsiOSmalloc(sizeof(iscsiHeader_t));
		if (NULL == message) {
			return NULL;
		}
	}
	message->forw = NULL;
	PRHEX(message);
	TRACE_EXIT;
	return message;
}


/* ************************************************ */


/*
* NAME: getRttStruct
*
* PURPOSE: return pointer to an RTT structure
*
* PARAMETERS: none
*
* RETURNS: pointer to rtt structure
*
* CALLED BY: handleRtt()
*
* NOTES:
*/
rttInfo_t *getRttStruct(void) {
	rttInfo_t *rtt;

	TRACE_ENTER;
	Qget(&freeRttQ, rtt, rttInfo_t);
	if (NULL == rtt) {
		rtt = iscsiOSmalloc(sizeof(rttInfo_t));
		if (NULL == rtt) {
			return NULL;
		}
	}
	rtt->forw = NULL;
	TRACE_EXIT;
	return rtt;
}



/* ************************************************ */

/*
* NAME: setTaskId
*
* PURPOSE: assign an iSCSI Task Tag to a command
*
* PARAMETERS:
*	cmd - pointer to pending command structure
*
* RETURNS: value of task identifier
*
* CALLED BY: sendCmd(), sendTaskManage(), etc
*
* NOTES:
* Generate a unique (initiator-wide) task id.
* For now, we use the (32-bit) address of the structure.
* This may change in future implementations.
* When the response arrives, we should check the validity of
* the Task id before depending on it to be a valid address.
*/
u32 setTaskId(void *cmd)
{
	u32 taskId;
	TRACE_ENTER;
	taskId = (u32) cmd;
	PRHEX(taskId);
	TRACE_EXIT;
	return taskId;
}



/*
* NAME: lookupTaskId
*
* PURPOSE: map iSCSI Task Tag to pointer of corresponding structure
*
* PARAMETERS:
*	si - pointer to session in which we should find the task id 
*	taskId - iSCSI Task Tag
*	flag - indicates COMMAND or TASK_MANAGEMENT, and whether to remove
*
* RETURNS: 
*	pointer to structuure corresponding to task tag
*	NULL - if not found
*
* CALLED BY: handleScsiResponse(), handleDataRead(), etc
*
* NOTES:
* Convert a taskId to a pointer to an internal data structure.
* Verify that address is valid.
* If found and flag specifies to remove, remove command from specified
* table. Return pointer to structure.
* If not found, return NULL.
*/
void *lookupTaskId(sessionInfo_t *si, u32 taskId, u32 flag)
{
	pendingCmd_t	*cmd;
	taskManageInfo_t	*tm;
	pendCmdTable_t	*pct;
	pendTaskManTable_t	*ptmt;
	void	*ptr;
	iscsiIrql_t irql;

	TRACE_ENTER;
	PRHEX(taskId);
	PRHEX(flag);
	ptr = (void *) taskId;
	if (ISCSI_TASKID_COMMAND & flag) {
		pct = &si->pendCmdTable;;
		iscsiOSlock(&pct->lock, &irql);
		for (cmd = pct->head->forw; cmd != pct->head; cmd = cmd->forw) {
			PRHEX(cmd);
			if (ptr == cmd) {
				if (ISCSI_TASKID_REMOVE & flag) {
					cmd->back->forw = cmd->forw;
					cmd->forw->back = cmd->back;
					cmd->forw = NULL;
					cmd->back = NULL;
				}
				iscsiOSunlock(&pct->lock, irql);
				TRACE_EXIT;
				return ptr;
			}
		}
		iscsiOSunlock(&pct->lock, irql);
	}
	else if (ISCSI_TASKID_TASK_MANGEMENT & flag) {
		ptmt = &si->pendTaskManTable;
		iscsiOSlock(&ptmt->lock, &irql);
		for (tm = ptmt->head->forw; tm != ptmt->head; tm = tm->forw) {
			PRHEX(tm);
			if (ptr == tm) {
				if (ISCSI_TASKID_REMOVE & flag) {
					tm->back->forw = tm->forw;
					tm->forw->back = tm->back;
					tm->forw = NULL;
					tm->back = NULL;
				}
				iscsiOSunlock(&ptmt->lock, irql);
				TRACE_EXIT;
				return ptr;
			}
		}
		iscsiOSunlock(&ptmt->lock, irql);
	}
	TRACE_EXIT;
	return NULL;
}


/* ************************************************ */



/*
* NAME: enterIntoTaskManTable
*
* PURPOSE: enter a Task Management operation into Task Management table
*
* PARAMETERS:
*	si - pointer to relevant session
*	tm - pointer to Task Management structure to save
*
* RETURNS: none
*
* CALLED BY: sendTaskMessage()
*
* NOTES:
*/
void enterIntoTaskManTable(sessionInfo_t *si, taskManageInfo_t *tm)
{
	iscsiIrql_t irql;

	TRACE_ENTER;
	iscsiOSlock(&si->pendTaskManTable.lock, &irql);
	tm->forw = si->pendTaskManTable.head;
	tm->back = si->pendTaskManTable.head->back;
	si->pendTaskManTable.head->back->forw = tm;
	si->pendTaskManTable.head->back = tm;
	iscsiOSunlock(&si->pendTaskManTable.lock, irql);
	TRACE_EXIT;
	return;
}



/*
* NAME: removeFromTaskManTable
*
* PURPOSE: remove specified Task Management request from table
*
* PARAMETERS:
*	si - pointer to relevant session
*	tm - pointer to Task Management structure to remove
*
* RETURNS: none
*
* CALLED BY: handleTaskManResponse()
*
* NOTES:
* xxx Do we have to check whether the tm structure is still in the table?
*/
void removeFromTaskManTable(sessionInfo_t *si, taskManageInfo_t *tm)
{
	iscsiIrql_t irql;

	TRACE_ENTER;
	iscsiOSlock(&si->pendTaskManTable.lock, &irql);
	tm->back->forw = tm->forw;
	tm->forw->back = tm->back;
	iscsiOSunlock(&si->pendTaskManTable.lock, irql);
	tm->forw = NULL;
	tm->back = NULL;
	TRACE_EXIT;
}




/* ************************************************ */

