
/*
*
* iscsi initiator driver
*
* Author: Kalman Meth, IBM Haifa Research Laboratory <meth@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* query.c
*
* Routines to query iscsi driver data structures.
*
* %I% %W% %D% %T%
*/

#include "iscsi_core_types.h"
#include "iscsi_core_query.h"
#include "iscsi_core_queue.h"
#include "iscsi_core_prot.h"




/*
* NAME: iscsiCoreQuerySession
*
* PURPOSE: Gather some information about a session
*
* PARAMETERS:
*	sessionHandle - handle (index) of session
*	qs - pointer to session query structure
*
* RETURNS:
*	0 - success
*	-1 - error
*
* CALLED BY: OS dependent ioctl function
*
* NOTES:
*/
s32 iscsiCoreQuerySession(u32 sessionHandle, querySession_t *qs)
{
	u32		count;
	sessionInfo_t	*si;
	iscsiIrql_t	irql;
	pendingCmd_t	*cmd;
	taskManageInfo_t	*tm;

	si = getSessionStruct(sessionHandle);
	if (NULL == si) {
		return -1;
	}

	qs->cmdRN = si->cmdRN;
	qs->expCmdRN = si->expCmdRN;
	qs->maxCmdRN = si->maxCmdRN;
	qs->numMaxConnections = si->numMaxConnections;
	qs->numActiveConnections = si->numActiveConnections;
	qs->isid = si->isid;
	qs->tsid = si->tsid;

	qs->nCmdQentries = si->cmdQ.count;
	qs->nRttQentries = si->rttQ.count;
	qs->nTmQentries = si->taskManQ.count;

	count = 0;
	iscsiOSlock(&si->pendCmdTable.lock, &irql);
	for (cmd = si->pendCmdTable.head->forw; cmd != si->pendCmdTable.head; cmd = cmd->forw) {
		count++;
	}
	iscsiOSunlock(&si->pendCmdTable.lock, irql);
	qs->nPendCmds = count;

	count = 0;
	iscsiOSlock(&si->pendTaskManTable.lock, &irql);
	for (tm = si->pendTaskManTable.head->forw; tm != si->pendTaskManTable.head; tm = tm->forw) {
		count++;
	}
	iscsiOSunlock(&si->pendTaskManTable.lock, irql);
	qs->nPendTms = count;

	return 0;
}

/*
* NAME: iscsiCoreGetSessionStats
*
* PURPOSE: return statistics for specified session
*
* PARAMETERS:
*	sessionHandle - handle (index) of session
*	stats - pointer to structure that contains the statistics
*
* RETURNS:
*	0 - success
*	-1 - error
*
* CALLED BY: OS dependent ioctl function
*
* NOTES:
*/
s32 iscsiCoreGetSessionStats(u32 sessionHandle, sessionStats_t *stats)
{
	sessionInfo_t	*si;

	si = getSessionStruct(sessionHandle);
	if (NULL == si) {
		return -1;
	}

	memcpy(stats, si->stats, sizeof(sessionStats_t));
	return 0;
}

/*
* NAME: iscsiCoreQueryConnection
*
* PURPOSE: Gather some information about a connection
*
* PARAMETERS:
*	sessionHandle - handle (index) of session
*	qc - pointer to connection query structure
*
* RETURNS:
*	0 - success
*	-1 - error
*
* CALLED BY: OS dependent ioctl function
*
* NOTES:
*/
s32 iscsiCoreQueryConnection(u32 sessionHandle, queryConnection_t *qc)
{
	sessionInfo_t	*si;
	connInfo_t	*ci;
	u32	index;

	si = getSessionStruct(sessionHandle);
	if (NULL == si) {
		return -1;
	}

	index = qc->connectionNumber;
	if (index > si->numMaxConnections) {
		return -1;
	}
	ci = &si->connections[index];

	qc->addr = ci->addr;
	qc->portNum = ci->portNum;
	qc->state = ci->state;
	return 0;
}
