
/*
*
* iscsi initiator driver
*
* Author: Kalman Meth, IBM Haifa Research Laboratory <meth@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* receive.c
*
* Handler to read and handle responses from a TCP connection.
*
* %I% %W% %D% %T%
*/

#include "iscsi_interfaces.h"
#include "iscsi_core_messages.h"
#include "iscsi_local_trace.h"
#include "iscsi_core_prot.h"



/*
* NAME: registerStatRN
*
* PURPOSE: keep track of status reference numbers
*
* PARAMETERS:
*	si - pointer to session structure
*	statRN - status reference number of received response
*
* RETURNS: none
*
* CALLED BY: HandleScsiResponse(), HandleDataRead(), HandleTaskResponse()
*
* NOTES:
* Not yet implemented. Not neeeded with a single connection.
*
* FUTURE ENHANCEMENTS:
* We have to keep a list of which statRNs were received,
* so that we can detect a missing statRN and figure out expStatRN.
*/
void registerStatRN(sessionInfo_t *si, u32 statRN)
{
	/* in first implementation, since we have only a single TCP connection, there is nothing to do here. */
}


/*
* NAME: eatUpBytes
*
* PURPOSE: read bytes from tcp stream and discard them
*
* PARAMETERS:
*	sock - pointer to connection information
*	length - number of bytes to consume from connection
*
* RETURNS:
*	0 - success
*	-1 - catastrophic error
*
* CALLED BY: HandleScsiResponse()
*
* NOTES:
*
* FUTURE ENHANCEMENTS
* Allocate a single buffer for all eatUpBytes operation.
* Always keep the largest buffer allocated.
*/
s32 eatUpBytes(iscsiSock_t *sock, u32 length)
{
	uchar *ptr;
	s32 rc;

	if (length > 0) {
		ptr = iscsiOSmalloc(length);
		if (NULL == ptr) {
			/* nothing we can do; shut down connection */
			return -1;
		}
		rc = iscsiOSreadFromConnection(sock, ptr, length, 0, 0);
		iscsiOSfree(ptr);
		if (rc != (s32) length) {
			return -1;
		}
	}
	return 0;
}


/*
* NAME: handleScsiResponse
*
* PURPOSE: perform processing of a response to a SCSI command
*
* PARAMETERS:
*	ci - pointer to connection information where response arrived
*	response - iSCSI header of received response
*
* RETURNS: return -1 only for catastrophic errors; otherwise return 0
*
* CALLED BY: readConnectionHandler()
*
* NOTES:
* Return -1 only in case we loose synchronization while reading from the connection.
*/
s32 handleScsiResponse(connInfo_t *ci, iscsiHeader_t *response)
{
	sessionInfo_t	*si = ci->sInfo;
	iscsiSock_t	*sock = ci->sock;
	pendingCmd_t	*cmd;
	u32 length;
	u32 taskId;
	s32 rc;
	s32 rc2 = 0;
	u32 statRN;
	u32 expCmdRN;
	u32 maxCmdRN;
	iscsiIrql_t irql;
	u32 diff = 0;
	u16 senseLen = 0;
	u16 responseLen;
	u8 status = 0;
#define SENSE_DATA_SIZE	128
	uchar	senseData[SENSE_DATA_SIZE];

	TRACE_ENTER;
	PRHEX(response);

	/* extract length of data beyond the header */
	length = ntohl(*(u32 *) &response->buf[LENGTH_BYTE]);
	PRINT(length);

	/* Reference number handling */
	statRN = ntohl(*(u32 *) &response->buf[STAT_RN_BYTE]);
	PRINT(statRN);
	registerStatRN(si, statRN);
	expCmdRN = ntohl(*(u32 *) &response->buf[EXP_CMD_RN_BYTE]);
	PRINT(expCmdRN);
	maxCmdRN = ntohl(*(u32 *) &response->buf[MAX_CMD_RN_BYTE]);
	PRINT(maxCmdRN);

	/* update RNs */
	iscsiOSlock(&si->lock, &irql);
	if (GT_WITH_WRAP(expCmdRN, si->expCmdRN)) {
		si->expCmdRN = expCmdRN;
	}
	if (GT_WITH_WRAP(maxCmdRN, si->maxCmdRN)) {
		si->maxCmdRN = maxCmdRN;
	}
	iscsiOSunlock(&si->lock, irql);

	/* extract task tag to find pendingCmd structure */
	taskId = *((u32 *) &response->buf[TASK_TAG_BYTE]);
	cmd = lookupTaskId(si, taskId, ISCSI_TASKID_COMMAND | ISCSI_TASKID_REMOVE);
	if (NULL == cmd) {
		/* command not found in table */
		ISCSI_INCREMENT_COUNTER(si, nBadTaskId);
		ISCSI_INCREMENT_COUNTER(si, nResponsesDiscarded);
		Qput(&freeMsgHdrQ, response, iscsiHeader_t);
		/* xxx log an error? */
		/* eat up any additional data in order to get to next header */
		rc = eatUpBytes(sock, length);
		return rc;
	}

	status = response->buf[STATUS_BYTE];
	PRINT(status);
#ifdef GATHER_STATISTICS
	if (status) {
		switch(status) {
			case ISCSI_SCSISTAT_CHECK_CONDITION:
				ISCSI_INCREMENT_COUNTER(si,nCheckConditionReceived);
				break;
			case ISCSI_SCSISTAT_BUSY:
				ISCSI_INCREMENT_COUNTER(si, nBusyReceived);
				break;
			default:
				ISCSI_INCREMENT_COUNTER(si, nOtherStatusReceived);
				break;
		}
	}
#endif /* GATHER_STATISTICS */

	responseLen = ntohs(*(u16 *)&response->buf[RESPONSE_LEN_BYTE]);
	PRINT(responseLen);
	if (responseLen) {
		u32 readLen = responseLen;

		if (responseLen > cmd->datalen) {
			readLen = cmd->datalen;
			diff = responseLen - readLen;
		}
		PRINT(readLen);
		if (readLen > 0) {
			rc = iscsiOSreadFromConnection(sock, cmd->data, readLen, 0, 0);
			if (rc != (s32) readLen) {
				/* unrecoverable error */
				/* complete command with error */
				rc2 = -1;
				goto finish;
			}
		}
		/* eat up rest of data that arrived beyond readLen */
		if (diff) {
			rc2 = eatUpBytes(sock, diff);
			if (-1 == rc2) {
				/* unrecoverable error */
				/* complete command with error */
				goto finish;
			}
		}
	}
	senseLen = ntohs(*(u16 *)&response->buf[SENSE_LEN_BYTE]);
	if (senseLen > 0)
	/* xxx if (ISCSI_SCSISTAT_CHECK_CONDITION == status) */
	{
		/* provide sense data, if present */
		void 	*ptr = senseData;
		u32	didMalloc = 0;
		PRINT(senseLen);
		if (senseLen > SENSE_DATA_SIZE) {
			ptr = iscsiOSmalloc(senseLen);
			if (NULL == ptr) {
				/* nothing we can do; shut down connection */
				rc2 = -1;
				goto finish;
			}
			didMalloc = 1;
		}
		PRHEX(ptr);
		rc = iscsiOSreadFromConnection(sock, ptr, senseLen, 0, 0);
		if (rc != (s32) senseLen) {
			/* unrecoverable error */
			/* complete command with error */
			if (didMalloc) {
				iscsiOSfree(ptr);
			}
			rc2 = -1;
			goto finish;
		}
		PRBUF(ptr, senseLen);
		PRINT(senseLen);
		PRINT(sizeof(cmd->scb->sense_buffer));
		PRINT(MIN(senseLen, sizeof(cmd->scb->sense_buffer)));
		iscsiOSputSense(cmd->scb, ptr, senseLen);
		if (didMalloc) {
			iscsiOSfree(ptr);
		}
	}
	if ((u32)(responseLen + senseLen) != length) {
		/* we have a problem */
		/* eat up remaining data */
		diff = length - (responseLen + senseLen);
		/* xxx what if length < (responseLen + senseLen) ??? */
		PRINT(diff);
		rc2 = eatUpBytes(sock, diff);
		if (-1 == rc2) {
			/* Note: command was completely processed. sense data, if present, was already set */
		}
	}

	/* xxx what do we do with residual count? */
	/* xxx do we ever need to check iSCSI status? */

	PRHEX(cmd);
	PRHEX(cmd->header);
	PRHEX(response);
	iscsiOScompleteCommand(cmd->scb, status);
	Qput(&freeMsgHdrQ, cmd->header, iscsiHeader_t);
	Qput(&freeMsgHdrQ, response, iscsiHeader_t);
	Qput(&freeCmdQ, cmd, pendingCmd_t);

	TRACE_EXIT;
	return rc2;

finish:
	/* if have unrecoverable error, set special sense data, in case we didn't yet do it */
	/* complete command with error */
	status = ISCSI_SCSISTAT_CHECK_CONDITION;
	/* LOGICAL UNIT COMMUNICATION ERROR (0x08, 0x00) */
	SET_SENSE(senseData, ISCSI_SCSISENSE_NOT_READY, 0x08, 0x00);
	iscsiOSputSense(cmd->scb, senseData, sizeof(senseData_t));
	iscsiOScompleteCommand(cmd->scb, status);
	ISCSI_INCREMENT_COUNTER(si, nResponsesDiscarded);
	Qput(&freeMsgHdrQ, cmd->header, iscsiHeader_t);
	Qput(&freeMsgHdrQ, response, iscsiHeader_t);
	Qput(&freeCmdQ, cmd, pendingCmd_t);

	TRACE_EXIT;
	return rc2;
}


/*
* NAME: handleDataRead
*
* PURPOSE: perform processing of incoming data packet
*
* PARAMETERS:
*	ci - pointer to connection information where response arrived
*	response - iSCSI header of received data packet
*
* RETURNS: return -1 only for catastrophic errors; otherwise return 0
*
* CALLED BY: readConnectionHandler()
*
* NOTES:
*/
s32 handleDataRead(connInfo_t *ci, iscsiHeader_t *response)
{
	sessionInfo_t	*si = ci->sInfo;
	iscsiSock_t	*sock = ci->sock;
	pendingCmd_t	*cmd;
	u32 length;
	u32 statRN;
	u32 expCmdRN;
	u32 maxCmdRN;
	iscsiIrql_t irql;
	u32 taskId;
	u32 statusValid = 0;
	u32 offset;
	u32 diff;
	u32 readLen;
	s32 rc;
	s32 rc2 = 0;
	u8 status = 0;

	TRACE_ENTER;

	/* extract length of data beyond the header */
	length = ntohl(*(u32 *) &response->buf[LENGTH_BYTE]);
	PRINT(length);

	/* Reference number handling */
	statRN = ntohl(*(u32 *) &response->buf[STAT_RN_BYTE]);
	PRINT(statRN);
	registerStatRN(si, statRN);
	expCmdRN = ntohl(*(u32 *) &response->buf[EXP_CMD_RN_BYTE]);
	PRINT(expCmdRN);
	maxCmdRN = ntohl(*(u32 *) &response->buf[MAX_CMD_RN_BYTE]);
	PRINT(maxCmdRN);

	/* update RNs */
	iscsiOSlock(&si->lock, &irql);
	if (GT_WITH_WRAP(expCmdRN, si->expCmdRN)) {
		si->expCmdRN = expCmdRN;
	}
	if (GT_WITH_WRAP(maxCmdRN, si->maxCmdRN)) {
		si->maxCmdRN = maxCmdRN;
	}
	iscsiOSunlock(&si->lock, irql);

	/* extract task tag to find pendingCmd structure */
	taskId = *((u32 *) &response->buf[TASK_TAG_BYTE]);
	cmd = lookupTaskId(si, taskId, ISCSI_TASKID_COMMAND);
	if (NULL == cmd) {
		/* command not found in table */
		/* xxx log an error? */
		ISCSI_INCREMENT_COUNTER(si, nBadTaskId);
		/* eat up any additional data in order to get to next header */
		Qput(&freeMsgHdrQ, response, iscsiHeader_t);
		if (length > 0) {
			rc = eatUpBytes(sock, length);
			if (-1 == rc) {
				return -1;
			}
		}
		return 0;
	}


	/* check for status */
	if (response->buf[FLAGS_BYTE] & ISCSI_STATUS_BIT) {
		/* command is complete */
		statusValid = 1;
		status = response->buf[STATUS_BYTE];
		PRINT(status);
		cmd = removeCmdFromTable(si, cmd);
		if (NULL == cmd) {
			/* xxx log an error? */
			Qput(&freeMsgHdrQ, response, iscsiHeader_t);
			return 0;
		}
		/* xxx what do we do with residual count? */
		/* xxx do we ever need to check iSCSI status? */
	}

	/* copy data to supplied buffer */
	/* verify that data fits into allocated buffer */
	offset = ntohl(*(u32 *) &response->buf[DATA_BUFFER_OFFSET_BYTE]);
	PRINT(length);
	PRINT(offset);
	PRINT(cmd->datalen);
	if (offset >= cmd->datalen) {
		readLen = 0;
		diff = length;
	}
	else if ((length + offset) > cmd->datalen) {
		readLen = cmd->datalen - offset;
		diff = length - readLen;
	}
	else {
		readLen = length;
		diff = 0;
	}
	PRINT(readLen);
	PRINT(diff);
	if (readLen) {
		rc = iscsiOSreadFromConnection(sock, cmd->data + offset, readLen, cmd->use_sg, offset);
		if (rc != (s32) readLen) {
			/* unrecoverable error */
			/* xxx complete command with error */
			rc2 = -1;
			goto finish;
		}
	}
	/* eat up rest of data that arrived beyond readLen */
	if (diff) {
		rc2 = eatUpBytes(sock, diff);
		if (-1 == rc2) {
			/* unrecoverable error */
			goto finish;
		}
	}

finish:
	Qput(&freeMsgHdrQ, response, iscsiHeader_t);
	if (statusValid) {
		/* complete the command */
		/* xxx handle case where status is not good due to failure above */
		iscsiOScompleteCommand(cmd->scb, status);
		Qput(&freeMsgHdrQ, cmd->header, iscsiHeader_t);
		Qput(&freeCmdQ, cmd, pendingCmd_t);
	}
	TRACE_EXIT;
	return rc2;
}


/*
* NAME: handleRtt
*
* PURPOSE: perform processing of an RTT packet
*
* PARAMETERS:
*	ci - pointer to connection information where response arrived
*	response - iSCSI header of received RTT packet
*
* RETURNS: return -1 only for catastrophic errors (lost sync); otherwise return 0
*
* CALLED BY: readConnectionHandler()
*
* NOTES:
*/
s32 handleRtt(connInfo_t *ci, iscsiHeader_t *response)
{
	sessionInfo_t	*si = ci->sInfo;
	iscsiSock_t	*sock = ci->sock;
	pendingCmd_t	*cmd;
	rttInfo_t	*rtt;
	u32 length;
	s32 rc;
	u32 taskId;
	u32 offset;
	u32 expCmdRN;
	u32 maxCmdRN;
	iscsiIrql_t irql;

	TRACE_ENTER;

	/* extract length of data beyond the header; should be 0 */
	length = ntohl(*(u32 *) &response->buf[LENGTH_BYTE]);
	PRINT(length);

	/* Reference number handling */
	expCmdRN = ntohl(*(u32 *) &response->buf[EXP_CMD_RN_BYTE]);
	PRINT(expCmdRN);
	maxCmdRN = ntohl(*(u32 *) &response->buf[MAX_CMD_RN_BYTE]);
	PRINT(maxCmdRN);

	/* update RNs */
	iscsiOSlock(&si->lock, &irql);
	if (GT_WITH_WRAP(expCmdRN, si->expCmdRN)) {
		si->expCmdRN = expCmdRN;
	}
	if (GT_WITH_WRAP(maxCmdRN, si->maxCmdRN)) {
		si->maxCmdRN = maxCmdRN;
	}
	iscsiOSunlock(&si->lock, irql);

	/* extract task tag to find pendingCmd structure */
	taskId = *((u32 *) &response->buf[TASK_TAG_BYTE]);
	cmd = lookupTaskId(si, taskId, ISCSI_TASKID_COMMAND);
	if (NULL == cmd) {
		/* command not found in table */
		/* xxx log an error? */
		ISCSI_INCREMENT_COUNTER(si, nBadTaskId);
		ISCSI_INCREMENT_COUNTER(si, nRttsRejected);
		/* eat up any additional data in order to get to next header */
		Qput(&freeMsgHdrQ, response, iscsiHeader_t);
		if (length > 0) {
			rc = eatUpBytes(sock, length);
			if (-1 == rc) {
				return -1;
			}
		}
		return 0;
	}

	offset = *((u32 *) &response->buf[RTT_BUFFER_OFFSET_BYTE]);
	if (offset > cmd->datalen) {
		/* RTT request beyond end of buffer */
		/* nothing we can do with this */
		/* xxx log error? */
		ISCSI_INCREMENT_COUNTER(si, nRttsRejected);
		Qput(&freeMsgHdrQ, response, iscsiHeader_t);
		return 0;
	}

	rtt = getRttStruct();
	if (NULL == rtt) {
		/* out of memory; nothing we can do */
		Qput(&freeMsgHdrQ, response, iscsiHeader_t);
		return 0;
	}
	rtt->pendingCmd = cmd;
	rtt->rttHeader = response;
	Qput(&si->rttQ, rtt, rttInfo_t);
	ISCSI_INCREMENT_COUNTER(si, nRttsQueued);
	iscsiOSlock(&si->lock, &irql);
	iscsiOSwakeup(&si->cmdQEvent);
	iscsiOSunlock(&si->lock, irql);

	TRACE_EXIT;
	return 0;
}


/*
* NAME: handleTaskManResponse
*
* PURPOSE: perform processing of a Task Management response
*
* PARAMETERS:
*	ci - pointer to connection information where response arrived
*	response - iSCSI header of received response
*
* RETURNS: return -1 only for catastrophic errors; otherwise return 0
*
* CALLED BY: readConnectionHandler()
*
* NOTES:
*/
s32 handleTaskManResponse(connInfo_t *ci, iscsiHeader_t *response)
{
	taskManageInfo_t	*tm;
	sessionInfo_t	*si = ci->sInfo;
	iscsiSock_t	*sock = ci->sock;
	pendingCmd_t	*cmd = NULL;
	u32 statRN;
	u32 expCmdRN;
	u32 maxCmdRN;
	iscsiIrql_t irql;
	u32 taskId;
	u32 length;
	s32 rc = 0;
	u8 status;

	TRACE_ENTER;

	/* extract length of data beyond the header */
	length = ntohl(*(u32 *) &response->buf[LENGTH_BYTE]);
	PRINT(length);

	/* Reference number handling */
	statRN = ntohl(*(u32 *) &response->buf[STAT_RN_BYTE]);
	PRINT(statRN);
	registerStatRN(si, statRN);
	expCmdRN = ntohl(*(u32 *) &response->buf[EXP_CMD_RN_BYTE]);
	PRINT(expCmdRN);
	maxCmdRN = ntohl(*(u32 *) &response->buf[MAX_CMD_RN_BYTE]);
	PRINT(maxCmdRN);

	/* update RNs */
	iscsiOSlock(&si->lock, &irql);
	if (GT_WITH_WRAP(expCmdRN, si->expCmdRN)) {
		si->expCmdRN = expCmdRN;
	}
	if (GT_WITH_WRAP(maxCmdRN, si->maxCmdRN)) {
		si->maxCmdRN = maxCmdRN;
	}
	iscsiOSunlock(&si->lock, irql);

	if (length > 0) {
		/* eat up any additional data in order to get to next header */
		/* xxx this shouldn't happen */
		/* xxx log an error? */
		rc = eatUpBytes(sock, length);
	}

	/* extract task tag to find pendTaskMan structure */
	taskId = *((u32 *) &response->buf[TASK_TAG_BYTE]);
	tm = lookupTaskId(si, taskId, ISCSI_TASKID_TASK_MANGEMENT | ISCSI_TASKID_REMOVE);
	PRHEX(taskId);
	PRHEX(tm);
	if (NULL == tm) {
		/* operation not found in table */
		PRMSG("XXX task management struct not found");
		/* xxx log an error? */
		ISCSI_INCREMENT_COUNTER(si, nBadTaskId);
		Qput(&freeMsgHdrQ, response, iscsiHeader_t);
		return rc;
	}

	tm->result = 0;

	status = response->buf[STATUS_BYTE];
	PRINT(status);

	Qput(&freeMsgHdrQ, response, iscsiHeader_t);

	PRINT(tm->operation);
	switch(tm->operation) {
	case ABORT_TASK:
		ISCSI_INCREMENT_COUNTER(si, nAbortsReceived);
		if (FUNCTION_COMPLETE == status) {
			tm->result = ISCSI_ABORT_SUCCESS;
			/* verify that command is still in table */
			/* we use here the fact that the task id is the same as the address of the pending command structure */
			cmd = lookupTaskId(si, (u32) tm->pendingCmd, ISCSI_TASKID_COMMAND | ISCSI_TASKID_REMOVE);
			if (NULL != cmd) {
				iscsiOScompleteCommandAbort(tm->oldScb);
				ISCSI_INCREMENT_COUNTER(si, nCmdsAborted);
				Qput(&freeMsgHdrQ, cmd->header, iscsiHeader_t);
				Qput(&freeCmdQ, cmd, pendingCmd_t);
			}
			else {
				tm->result = ISCSI_ABORT_CMD_NOT_FOUND;
			}
		}
		else {
			tm->result = ISCSI_ABORT_ERROR;
		}
		break;
	case LOGICAL_UNIT_RESET:
		ISCSI_INCREMENT_COUNTER(si, nResetsReceived);
		if (FUNCTION_COMPLETE == status) {
			/* cancel all pending commands for specified device */
			removeCmdsResetLun(si, tm->lun);
			tm->result = ISCSI_RESET_SUCCESS;
		}
		else {
			tm->result = ISCSI_RESET_ERROR;
		}
		break;
	case TARGET_RESET:
		ISCSI_INCREMENT_COUNTER(si, nResetsReceived);
		if (FUNCTION_COMPLETE == status) {
			tm->result = ISCSI_RESET_SUCCESS;
		}
		else {
			tm->result = ISCSI_RESET_ERROR;
		}
		/*
		In any case, call to breakTcpConnectionsAndEmptyQs()
		to break all connections and cancel all items in Qs.
		*/
		breakTcpConnectionsAndEmptyQs(si);
		break;
	default:
		PRMSG("unhandled task management response");
		break;
	}

	finishTaskManOperation(si, tm);

	TRACE_EXIT;
	return rc;
}


/*
* NAME: readConnectionHandler
*
* PURPOSE: thread to read from an iscsi connection
*
* PARAMETERS:
*	arg - pointer to connection information
*
* RETURNS: none
*
* CREATED BY: makeTcpConnections()
*
* NOTES:
*/
iscsiThreadEntry readConnectionHandler(void *arg)
{
	connInfo_t	*ci = arg;
	sessionInfo_t	*si = ci->sInfo;
	iscsiHeader_t	*response;
	iscsiSock_t	*sock = ci->sock;
	s32		rc;
	iscsiIrql_t	irql;
	u32		length;

	TRACE_ENTER;
	PRHEX(ci);

	while(1) {
		response = getMessageHeader();
		PRHEX(response);
		if (NULL == response) {
			/* panic; exit thread and clean up */
			break;
		}
		rc = iscsiOSreadFromConnection(sock, response->buf, ISCSI_HEADER_LEN, 0, 0);
		PRINT(rc);
		if (rc != ISCSI_HEADER_LEN) {
			/* panic; exit thread and clean up */
			Qput(&freeMsgHdrQ, response, iscsiHeader_t);
			break;
		}

		/* xxx perform common field analysis here */

		PRMSG("before switch");
		PRHEX(ISCSI_OPCODE(response));
		switch(ISCSI_OPCODE(response)) {
		case ISCSI_SCSI_RESPONSE:
			ISCSI_INCREMENT_COUNTER(si, nCmdsReceived);
			rc = handleScsiResponse(ci, response);
			break;
		case ISCSI_RTT:
			ISCSI_INCREMENT_COUNTER(si, nRttsReceived);
			rc = handleRtt(ci, response);
			break;
		case ISCSI_DATA_READ:
			ISCSI_INCREMENT_COUNTER(si, nDataReceived);
			rc = handleDataRead(ci, response);
			break;
		case ISCSI_TASK_RESPONSE:
			ISCSI_INCREMENT_COUNTER(si, nTMsReceived);
			rc = handleTaskManResponse(ci, response);
			break;
		case ISCSI_BAD_OPCODE:
			/* xxx log the information of bad opcode? */
			ISCSI_INCREMENT_COUNTER(si, nRejectOpCodeReceived);
			ISCSI_INCREMENT_COUNTER(si, nResponsesDiscarded);
			Qput(&freeMsgHdrQ, response, iscsiHeader_t);
			/* eat up data specified in Length field */
			length = ntohl(*(u32 *) &response->buf[LENGTH_BYTE]);
			rc = eatUpBytes(sock, length);
			break;
		case ISCSI_NOP_RESPONSE:
		case ISCSI_TEXT_RESPONSE:
		case ISCSI_ASYNC_EVENT:
		case ISCSI_PING_RESPONSE:
		case ISCSI_MAP_RESPONSE:
			/* response type not supported in this version of driver */
			ISCSI_INCREMENT_COUNTER(si, nResponsesDiscarded);
			Qput(&freeMsgHdrQ, response, iscsiHeader_t);
			/* eat up data specified in Length field */
			length = ntohl(*(u32 *) &response->buf[LENGTH_BYTE]);
			rc = eatUpBytes(sock, length);
			/* xxx update RNs? */
			break;
		default:
			/* opcode not recognized */
			ISCSI_INCREMENT_COUNTER(si, nBadOpCodeReceived);
			ISCSI_INCREMENT_COUNTER(si, nResponsesDiscarded);
			/* assume we have a TCP error; close connection */
			/* xxx log an error? */
			Qput(&freeMsgHdrQ, response, iscsiHeader_t);
			rc = -1;
			break;
		}
		PRINT(rc);
		if (rc < 0) {
			/* fatal error */
			break;
		}
	}

	/* if exit main loop, must clean up and exit */
	PRMSG("exited main loop");

	/* clean up pending commands of connection */
	removeCmdsBadConnection(si, ci);
	removeTaskManBadConnection(si, ci);

	PRINT(ci - si->connections);
	PRINT(si->numActiveConnections);
	iscsiOSlock(&si->lock, &irql);
	ci->state = CONNECTION_NOT_CONNECTED;
	/*
	Remove connection entry from sessionInfo by copying last
	active connection entry to overwrite this entry.
	*/
	si->numActiveConnections--;
	si->activeConnections[ci - si->connections] = si->activeConnections[si->numActiveConnections];

	/*
	if this was last connection of session,
	then we have completed the RESET operation
	*/
	if (!si->numActiveConnections) {
		si->flags &= ~SESSION_RESETTING;
	}

	/* wakeup up thread in freeSessionResources() */
	if (si->flags & SESSION_MUST_EXIT) {
		iscsiOSwakeup(&si->exitEvent);
	}

	iscsiOSunlock(&si->lock, irql);
	PRINT(si->numActiveConnections);

	TRACE_EXIT;
	return threadReturn;
}



