
/*
*
* iscsi initiator driver
*
* Author: Kalman Meth, IBM Haifa Research Laboratory <meth@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* send.c
*
* Routines and data structures needed for command queueing
* and to send out iscsi messages.
*	iscsiCoreEnterCmdInQ()
*
* See task.c for for additional queue manipulations needed for
* task management operations.
*
* %I% %W% %D% %T%
*/

#include "iscsi_interfaces.h"
#include "iscsi_core_messages.h"
#include "iscsi_local_trace.h"
#include "iscsi_core_queue.h"
#include "iscsi_core_prot.h"



/* queues of free structures */
Q_t freeCmdQ;		/* command structures */
Q_t freeTaskManQ;	/* Task Management structures */
Q_t freeRttQ;		/* RTT structures */
Q_t freeMsgHdrQ;	/* message header structures */




/*
* NAME: initGlobalQs
*
* PURPOSE: initialize global queues
*
* PARAMETERS: none
*
* RETURNS: none
*
* CALLED BY: iscsiCoreInitData()
*
* NOTES:
*/
void initGlobalQs(void) {
	pendingCmd_t *cmd;
	taskManageInfo_t *tm;
	u32 i;

	Qinit(&freeCmdQ);
	Qinit(&freeTaskManQ);
	Qinit(&freeRttQ);
	Qinit(&freeMsgHdrQ);

	/* allocate a bunch of structures in freeCmdQ and freeTaskManQ */
	/* xxx
	change this to a single giant block allocation, and divide up
	the memory chunk into individual structures.
	*/
	for (i = 0; i < ISCSI_CAN_QUEUE; i++) {
		cmd = iscsiOSmalloc(sizeof(pendingCmd_t));
		if (NULL == cmd) {
			return;
		}
		memset(cmd, 0, sizeof(pendingCmd_t));
		Qput(&freeCmdQ, cmd, pendingCmd_t);
	}
	/*
	We need enough Task Management structures to accomodate pending
	Abort operations on all pending commands, plus some additional
	Task Management operations with which to unjam the driver.
	*/
	for (i = 0; i < ISCSI_MAX_TASK_MANAGEMENT; i++) {
		tm = iscsiOSmalloc(sizeof(taskManageInfo_t));
		if (NULL == tm) {
			return;
		}
		memset(tm, 0, sizeof(taskManageInfo_t));
		iscsiOSeventInit(&tm->event);
		Qput(&freeTaskManQ, tm, taskManageInfo_t);
	}
}



/*
* NAME: cleanupGlobalQs
*
* PURPOSE: free up all resources used by global queues
*
* PARAMETERS: none
*
* RETURNS: none
*
* CALLED BY: iscsiCoreCleanupData()
*
* NOTES:
*/
void cleanupGlobalQs(void) {
	pendingCmd_t		*cmd;
	taskManageInfo_t	*tm;
	rttInfo_t		*rtt;
	iscsiHeader_t		*message;

	TRACE_ENTER;
	while (1) {
		Qget(&freeRttQ, rtt, rttInfo_t);
		if (NULL == rtt) {
			break;
		}
		PRHEX(rtt);
		iscsiOSfree(rtt);
	}
	while (1) {
		Qget(&freeMsgHdrQ, message, iscsiHeader_t);
		if (NULL == message) {
			break;
		}
		PRHEX(message);
		iscsiOSfree(message);
	}
	while (1) {
		Qget(&freeTaskManQ, tm, taskManageInfo_t);
		if (NULL == tm) {
			break;
		}
		iscsiOSeventDestroy(&tm->event);
		PRHEX(tm);
		iscsiOSfree(tm);
	}
	while (1) {
		Qget(&freeCmdQ, cmd, pendingCmd_t);
		if (NULL == cmd) {
			break;
		}
		PRHEX(cmd);
		iscsiOSfree(cmd);
	}
	TRACE_EXIT;
	return;
}



/*
* NAME: iscsiCoreEnterCmdInQ
*
* PURPOSE: enters specified SCSI command into specified session's command queue
*
* PARAMETERS:
*	scb - pointer to OS dependent SCSI Control Block of enqueued command
*	cdb - pointer to SCSI Command Descriptor Block
*	cdbLen - length of cdb
*	sessionHandle - index of session (target) for which command is intended
*	lun - (8-byte) Logical Unit Number (in SAM-2 format) for which command is intended
*	data - pointer to data buffer for command
*	datalen - length of data buffer
*	flags - DATA_IN, DATA_OUT, etc
*
* RETURNS: 0 for success, -1 for error
*
* CALLED BY: OS dependent routine
*
* NOTES:
*	All fields must be provided by caller.
*	The same data pointer parameter is used for both READ and WRITE.
*
* FUTURE ENHANCEMENTS:
*	Have different return codes for different types of failures
*	such as Q full, bad session, no memory, etc
*/
s32 iscsiCoreEnterCmdInQ(SCB_t *scb, void *cdb, u32 cdbLen,
			u32 sessionHandle, iscsiLUN_t lun, void *data,
			u32 datalen, u32 use_sg, u32 flags)
{
	pendingCmd_t *cmd;
	sessionInfo_t *si;
	iscsiIrql_t irql;

	TRACE_ENTER;

	/* verify that session is still valid */
	si = getSessionStruct(sessionHandle);
	if (NULL == si) {
		return -1;
	}
	PRHEX(si);

	/* if Q is full, return immediately with indication */
	if (si->cmdQ.count > (u32) (si->maxCmdRN - si->cmdRN)) {
		/* too many pending commands; return BUSY */
		ISCSI_INCREMENT_COUNTER(si, nCmdsBusy);
		return -1;
	}

	/* if session is being reset, do not enter into Q */
	if (si->flags & SESSION_RESETTING) {
		ISCSI_INCREMENT_COUNTER(si, nCmdsBusy);
		return -1;
	}

	/*
	try to take a structure from the free Q.
	If empty, return BUSY.
	*/
	PRHEX(freeCmdQ.head);
	PRHEX(freeCmdQ.tail);
	Qget(&freeCmdQ, cmd, pendingCmd_t);
	PRHEX(cmd);
	if (NULL == cmd) {
		/* too many pending commands; return BUSY */
		PRMSG("too many commands in Q");
		ISCSI_INCREMENT_COUNTER(si, nCmdsBusy);
		return -1;
	}

	/* fill in relevant fields */
	cmd->scb = scb;
	cmd->cdb = cdb;
	cmd->cdbLen = cdbLen;
	memcpy(cmd->lun, lun, sizeof(iscsiLUN_t));
	PRHEX(lun[1]);
	PRHEX(cmd->lun[1]);
	cmd->sInfo = si;
	cmd->data = data;
	cmd->datalen = datalen;
	cmd->use_sg = use_sg;
	cmd->flags = flags;

	/* place item in Q */
	Qput(&si->cmdQ, cmd, pendingCmd_t);
	ISCSI_INCREMENT_COUNTER(si, nCmdsQueued);

	/* wake up the handler */
	iscsiOSlock(&si->lock, &irql);
	iscsiOSwakeup(&si->cmdQEvent);
	iscsiOSunlock(&si->lock, irql);

	TRACE_EXIT;
	return 0;
}


/*
* NAME: completeTaskManLocal
*
* PURPOSE: complete a Task Mangement operation when cannot send message to target
*
* PARAMETERS:
*	si - pointer to session structure
*	tm - pointer to Task Management structure
*
* RETURNS: none
*
* CALLED BY: sendTaskMessage()
*
* NOTES:
* This function is called when we cannot send a message to the target.
* Do whatever we can to clean up locally.
* There are no read threads running for this session.
*/
void completeTaskManLocal(sessionInfo_t *si, taskManageInfo_t *tm)
{
	pendingCmd_t	*cmd = tm->pendingCmd;

	TRACE_ENTER;
	switch (tm->operation) {
	case ABORT_TASK:
		/* remove command from Pending Command Table */
		/* is there a race condition here? */
		if (NULL != cmd) {
			cmd = removeCmdFromTable(si, cmd);
			if (NULL == cmd) {
				tm->result = ISCSI_ABORT_CMD_NOT_FOUND;
			}
			else {
				iscsiOScompleteCommandAbort(cmd->scb);
				ISCSI_INCREMENT_COUNTER(si, nCmdsAborted);
				Qput(&freeMsgHdrQ, cmd->header, iscsiHeader_t);
				Qput(&freeCmdQ, cmd, pendingCmd_t);
				tm->result = ISCSI_ABORT_SUCCESS;
			}
		}
		else {
			tm->result = ISCSI_ABORT_CMD_NOT_FOUND;
		}
		break;
	case LOGICAL_UNIT_RESET:
		tm->result = ISCSI_RESET_ERROR;
		/* clean up locally all pending commands of specified LUN */
		removeCmdsResetLun(si, tm->lun);
	    	break;
	case TARGET_RESET:
		tm->result = ISCSI_RESET_SUCCESS;
		/* clean up locally the session connections and pending commands */
		breakTcpConnectionsAndEmptyQs(si);
		break;
	default:
		tm->result = ISCSI_RESET_ERROR;
	    	break;
	}
	finishTaskManOperation(si, tm);
	TRACE_EXIT;
}


/*
* NAME: sendTaskManage
*
* PURPOSE: send specified Task Management operation to target
*
* PARAMETERS:
*	si - pointer to session structure
*	tm - pointer to Task Management structure
*
* RETURNS: none
*
* CALLED BY: commandQHandler()
*
* NOTES:
* Always send Task Management message over first connection.
* Special handling for BUS_RESET, since we do it entirely locally.
*
* FUTURE ENHANCEMENTS:
* Dedicate first connection for Task Management operations
* to ensure that a Task Management operation is never held up
* because of a full pipe.
*/
void sendTaskManage(sessionInfo_t *si, taskManageInfo_t *tm)
{
	iscsiHeader_t	*message;
	pendingCmd_t	*cmd = tm->pendingCmd;
	iscsiIrql_t	irql;
	connInfo_t	*ci;
	s32		rc;

	TRACE_ENTER;
	PRHEX(si);
	PRHEX(tm);
	PRHEX(cmd);

	if (!si->numActiveConnections) {
		completeTaskManLocal(si, tm);
		return;
	}

	/* create task tag  */
	tm->taskId = setTaskId(tm);

	/* construct iscsi message and send to target */
	message = getMessageHeader();
	PRHEX(message);
	if (NULL == message) {
		/* comlete Task Management operation with an error */
		completeTaskManLocal(si, tm);
		return;
	}

	memset(message, 0, sizeof(iscsiHeader_t));
	message->buf[OP_CODE_BYTE] = ISCSI_TASK_MANAGEMENT;
	message->buf[FUNCTION_BYTE] = tm->operation;
	* (u32 *) &message->buf[LENGTH_BYTE] = 0;
	* (u32 *) &message->buf[TASK_TAG_BYTE] = tm->taskId;
	* (u32 *) &message->buf[CMD_RN_BYTE] = 0;
	* (u32 *) &message->buf[EXP_STAT_RN_BYTE] = htonl(si->expStatRN);

	iscsiOSlock(&si->lock, &irql);
	* (u32 *) &message->buf[CMD_RN_BYTE] = htonl(si->cmdRN);
	si->cmdRN++;
	if (!si->cmdRN) {
		si->cmdRN = 1;
	}
	iscsiOSunlock(&si->lock, irql);

	PRINT(tm->operation);
	switch(tm->operation) {
	case ABORT_TASK:
		* (u32 *) &message->buf[REFERENCED_TAG_BYTE] = * (u32 *) &cmd->header->buf[TASK_TAG_BYTE];
		memcpy(&message->buf[LUN_FIRST_BYTE], &cmd->header->buf[LUN_FIRST_BYTE], sizeof(iscsiLUN_t));
		ISCSI_INCREMENT_COUNTER(si, nAbortsSent);
		break;
	case LOGICAL_UNIT_RESET:
		memcpy(&message->buf[LUN_FIRST_BYTE], tm->lun, sizeof(iscsiLUN_t));
		ISCSI_INCREMENT_COUNTER(si, nResetsSent);
		break;
	case TARGET_RESET:
		/* nothing more to do */
		ISCSI_INCREMENT_COUNTER(si, nResetsSent);
		break;
	default:
		/* xxx not handled; log error */
		break;
	}

	tm->header = message;

	/* place into list of pending Task Management operations */
	enterIntoTaskManTable(si, tm);

	/* send the message over first of session's TCP connections */
	ci = si->activeConnections[0];
	tm->connInfo = ci;
	rc = iscsiOSwriteToConnection(ci->sock, message->buf, ISCSI_HEADER_LEN, 0, 0);
	if (rc != ISCSI_HEADER_LEN) {
		ci->state = CONNECTION_ERROR;
		removeFromTaskManTable(si, tm);
		/* perform locally */
		completeTaskManLocal(si, tm);
		Qput(&freeMsgHdrQ, message, iscsiHeader_t);
	}

	TRACE_EXIT;
	return;
}


/*
* NAME: sendRttData
*
* PURPOSE: send data buffer as specified in RTT request
*
* PARAMETERS:
*	si - pointer to session structure
*	rtt - pointer to RTT structure
*
* RETURNS: none
*
* CALLED BY: commandQHandler()
*
* NOTES:
*/
void sendRttData(sessionInfo_t *si, rttInfo_t *rtt) {
	pendingCmd_t	*cmd = rtt->pendingCmd;
	iscsiHeader_t	*message;
	iscsiIrql_t	irql;
	u32 offset;
	u32 length;
	s32 rc;

	TRACE_ENTER;
	PRHEX(si);
	PRHEX(rtt);
	PRHEX(cmd);

	if (!si->numActiveConnections) {
		/* nothing more we can do */
		Qput(&freeMsgHdrQ, rtt->rttHeader, iscsiHeader_t);
		Qput(&freeRttQ, rtt, rttInfo_t);
		/*
		command will be completed with an error when the read
		thread of the bad connection exits.
		*/
		return;
	}

	/* construct iscsi message and send to target */
	message = getMessageHeader();
	if (NULL == message) {
		/* nothing more we can do; panic? */
		Qput(&freeMsgHdrQ, rtt->rttHeader, iscsiHeader_t);
		Qput(&freeRttQ, rtt, rttInfo_t);
		return;
	}

	length = ntohl(* (u32 *) &rtt->rttHeader->buf[DESIRED_DATA_XFER_LEN]);
	offset = ntohl(* (u32 *) &rtt->rttHeader->buf[RTT_BUFFER_OFFSET_BYTE]);
	PRINT(length);
	PRINT(offset);
	/* we already checked earlier that (offset < cmd->datalen) */
	if ((offset + length) > cmd->datalen) {
		length = cmd->datalen - offset;
		PRINT(length);
	}

	memset(message, 0, sizeof(iscsiHeader_t));
	message->buf[OP_CODE_BYTE] = ISCSI_DATA_WRITE;

	* (u32 *) &message->buf[LENGTH_BYTE] = htonl(length);
	* (u32 *) &message->buf[DATA_BUFFER_OFFSET_BYTE] = htonl(offset);
	* (u32 *) &message->buf[DATA_TRANSFER_TAG] = * (u32 *) &rtt->rttHeader->buf[TARGET_TRANSFER_TAG];
	* (u32 *) &message->buf[TASK_TAG_BYTE] = cmd->taskId;
	* (u32 *) &message->buf[EXP_STAT_RN_BYTE] = htonl(si->expStatRN);
	iscsiOSlock(&si->lock, &irql);
	* (u32 *) &message->buf[CMD_RN_BYTE] = htonl(si->cmdRN);
	si->cmdRN++;
	if (!si->cmdRN) {
		si->cmdRN = 1;
	}
	iscsiOSunlock(&si->lock, irql);


#ifdef Windows

	//send the header and data together via iscsiOSwriteToConnection2
	rc = iscsiOSwriteToConnection2(cmd->connInfo->sock, message->buf, ISCSI_HEADER_LEN, cmd->data + offset, length);
	if (rc != ISCSI_HEADER_LEN + (s32)length)
	{
		cmd->connInfo->state = CONNECTION_ERROR;
	}
	else
	{
		ISCSI_INCREMENT_COUNTER(si, nDataSent);
	}

#else

	/* send the message over the TCP connection */
	rc = iscsiOSwriteToConnection(cmd->connInfo->sock, message->buf, ISCSI_HEADER_LEN, 0, 0);
	if (ISCSI_HEADER_LEN != rc) {
		cmd->connInfo->state = CONNECTION_ERROR;
	}
	else {
		rc = iscsiOSwriteToConnection(cmd->connInfo->sock, cmd->data + offset, length, cmd->use_sg, offset);
		if (rc != (s32) length) {
			cmd->connInfo->state = CONNECTION_ERROR;
		}
		else {
			ISCSI_INCREMENT_COUNTER(si, nDataSent);
		}
	}

#endif


	Qput(&freeMsgHdrQ, message, iscsiHeader_t);
	Qput(&freeMsgHdrQ, rtt->rttHeader, iscsiHeader_t);
	Qput(&freeRttQ, rtt, rttInfo_t);

	TRACE_EXIT;
	return;
}


/*
* NAME: sendCmd
*
* PURPOSE: send specified command to target
*
* PARAMETERS:
*	si - pointer to session structure
*	cmd - pointer to command information
*
* RETURNS: none
*
* CALLED BY: commandQHandler()
*
* NOTES:
* This function runs in the command handler thread.
* There are no other threads that might enter this function.
* We therefore do not need to hold the session lock for most of what
* we do in this funciton; only where there might be interaction with
* the read threads.
*/
void sendCmd(sessionInfo_t *si, pendingCmd_t *cmd)
{
	iscsiHeader_t *message;
	senseData_t senseData;
	int sendData = 0;
	iscsiIrql_t irql;
	s32 rc;

	TRACE_ENTER;
	PRHEX(si);
	PRHEX(cmd);

	if (!si->numActiveConnections) {
		/* complete command with error indication */
		/* xxx let's add here iscsiOScompleteCommandNoConnection() */
		/* LOGICAL UNIT COMMUNICATION ERROR (0x08, 0x00) */
#ifndef Windows
		SET_SENSE(senseData, ISCSI_SCSISENSE_NOT_READY, 0x08, 0x00);
		PRMSG("before iscsiOSputSense");
		iscsiOSputSense(cmd->scb, senseData, sizeof(senseData_t));
		PRMSG("before iscsiOScompleteCommand");
		iscsiOScompleteCommand(cmd->scb, ISCSI_SCSISTAT_CHECK_CONDITION);
#else
		iscsiOScompleteCommand(cmd->scb, 0xAB);
#endif
		/* return resources of message to pool */
		Qput(&freeCmdQ, cmd, pendingCmd_t);
		ISCSI_INCREMENT_COUNTER(si, nCmdsNoConnection);
		return;
	}

	if (GT_WITH_WRAP(si->cmdRN, si->maxCmdRN)) {
		/* too many pending commands; return BUSY */
		PRMSG("too many commands in Q");
		iscsiOScompleteCommand(cmd->scb, ISCSI_SCSISTAT_BUSY);
		Qput(&freeCmdQ, cmd, pendingCmd_t);
		ISCSI_INCREMENT_COUNTER(si, nCmdsBusy);
		return;
	}

	/* construct iscsi message and send to target */
	message = getMessageHeader();
	if (NULL == message) {
		/* complete the command with an error indication */
		iscsiOScompleteCommand(cmd->scb, ISCSI_SCSISTAT_BUSY);
		Qput(&freeCmdQ, cmd, pendingCmd_t);
		return;
	}
	PRHEX(message);
	PRHEX(cmd->flags);
	PRINT(cmd->datalen);

	memset(message, 0, sizeof(iscsiHeader_t));
	message->buf[OP_CODE_BYTE] = ISCSI_SCSI_COMMAND;

	/* determine if we have parameters */
	if (cmd->flags & ISCSI_DATA_OUT) {
		/* for small enough WRITEs, send as immediate data */
		if (cmd->datalen <= RTT_THRESHOLD) {
			* (u32 *) &message->buf[LENGTH_BYTE] = htonl(cmd->datalen);
			sendData = 1;
		}
		* (u32 *) &message->buf[EXPECTED_DATA_XFER_LEN] = htonl(cmd->datalen);
	}
	if (cmd->flags & ISCSI_DATA_IN) {
		/* set the R bit */
		message->buf[FLAGS_BYTE] |= R_BIT;
		* (u32 *) &message->buf[EXPECTED_DATA_XFER_LEN] = htonl(cmd->datalen);
	}

	PRHEX(cmd->lun[1]);
	memcpy(&message->buf[LUN_FIRST_BYTE], cmd->lun, sizeof(iscsiLUN_t));
	PRINT(si->cmdRN);
	PRINT(si->numActiveConnections);
	iscsiOSlock(&si->lock, &irql);
	cmd->connInfo = si->activeConnections[si->cmdRN % si->numActiveConnections];
	* (u32 *) &message->buf[CMD_RN_BYTE] = htonl(si->cmdRN);
	si->cmdRN++;
	if (!si->cmdRN) {
		si->cmdRN = 1;
	}
	* (u32 *) &message->buf[EXP_STAT_RN_BYTE] = htonl(si->expStatRN);
	iscsiOSunlock(&si->lock, irql);

	memcpy(&message->buf[CDB_FIRST_BYTE], cmd->cdb, cmd->cdbLen);
	cmd->header = message;

	/* create task tag  */
	cmd->taskId = setTaskId(cmd);
	PRHEX(cmd->taskId);
	*((u32 *) &message->buf[TASK_TAG_BYTE]) = cmd->taskId;

	/* place message in table of messages that were sent which are awaiting responses */
	enterCmdIntoTable(si, cmd);

	/* send the message over the TCP connection */
	PRHEX(message->buf);

	//the header we dont send in scatter gather
#ifdef Windows

	if (sendData) //send the header and data together via iscsiOSwriteToConnection2
	{
		rc = iscsiOSwriteToConnection2(cmd->connInfo->sock, message->buf, ISCSI_HEADER_LEN, cmd->data, cmd->datalen);
		if (rc != ISCSI_HEADER_LEN + (s32)cmd->datalen)
		{
			cmd->connInfo->state = CONNECTION_ERROR;
		}
		ISCSI_INCREMENT_COUNTER(si, nCmdsSent);
	}

	else //there is no data, so just send the header via iscsiOSwriteToConnection
	{
		rc = iscsiOSwriteToConnection(cmd->connInfo->sock, message->buf, ISCSI_HEADER_LEN, 0, 0);
		if (rc != ISCSI_HEADER_LEN)
		{
			cmd->connInfo->state = CONNECTION_ERROR;
		}
	}
	
#else

	rc = iscsiOSwriteToConnection(cmd->connInfo->sock, message->buf, ISCSI_HEADER_LEN, 0, 0);
	if (ISCSI_HEADER_LEN != rc) {
		cmd->connInfo->state = CONNECTION_ERROR;
	}
	else {
		if (sendData) {
			PRHEX(cmd->data);
			rc = iscsiOSwriteToConnection(cmd->connInfo->sock, cmd->data, cmd->datalen, cmd->use_sg, 0);
			if (rc != (s32) cmd->datalen) {
				cmd->connInfo->state = CONNECTION_ERROR;
			}
		}
		ISCSI_INCREMENT_COUNTER(si, nCmdsSent);
	}

#endif

	TRACE_EXIT;
	return;
}




/*
* NAME: cmdQHandler
*
* PURPOSE: process commands, etc, in a session's queues
*
* PARAMETERS:
*	arg - pointer to session structure
*
* RETURNS: none
*
* CREATED BY: iscsiCoreCreateSession()
*
* NOTES:
*/
iscsiThreadEntry cmdQHandler(void *arg)
{
	sessionInfo_t		*si = arg;
	pendingCmd_t		*cmd;
	taskManageInfo_t	*tm;
	rttInfo_t		*rtt;
	iscsiIrql_t		irql;

	TRACE_ENTER;
	PRHEX(si);

	do {
		PRMSG("top of the loop");

		/* if session has no active connections, make connections */
		PRINT(si->numActiveConnections);
		PRHEX(si->flags);
		if ((!si->numActiveConnections)  && (!(si->flags & (SESSION_MUST_EXIT | SESSION_RESETTING)))) {
			u32 rc;
			rc = makeTcpConnections(si);
			if (!rc) {
				/*
				If no connections, we will get an error
				when we try to send a message.
				At that time we will complete the
				request with an error indication.
				In any case, process next command.
				*/
			}
		}
		Qget(&si->taskManQ, tm, taskManageInfo_t);
		PRHEX(tm);
		if (NULL != tm) {
			sendTaskManage(si, tm);
			continue;
		}
		Qget(&si->rttQ, rtt, rttInfo_t);
		PRHEX(rtt);
		if (NULL != rtt) {
			sendRttData(si, rtt);
			continue;
		}
		Qget(&si->cmdQ, cmd, pendingCmd_t);
		PRHEX(cmd);
		if (NULL != cmd) {
			sendCmd(si, cmd);
			continue;
		}

		/* nothing in any Q; go to sleep */
		/*
		We get the session lock and release
		the lock atomically with going to sleep in order to
		avoid a race condition with the wakeup call.
		*/
		iscsiOSlock(&si->lock, &irql);
		if (!(si->flags & SESSION_MUST_EXIT)) {
			iscsiOSunlockAndSleepAndLock(&si->lock, &si->cmdQEvent, &irql);
		}
		iscsiOSunlock(&si->lock, irql);
	} while (!(si->flags & SESSION_MUST_EXIT));

	/* session's data structures are cleaned up in iscsiCoreDeleteSession() */
	/* indicate that thread has exited by setting threadId to 0 */
	iscsiOSlock(&si->lock, &irql);
	si->pendingCmdThreadId = 0;

	/* wakeup thread in iscsiCoreDeleteSession() */
	iscsiOSwakeup(&si->exitEvent);
	iscsiOSunlock(&si->lock, irql);

	TRACE_EXIT;
	return threadReturn;
}

