
/*
*
* iscsi initiator driver
*
* Author: Kalman Meth, IBM Haifa Research Laboratory <meth@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* session.c
*
* Session creation and deletion routines.
*	iscsiCoreCreateSession()
*	iscsiCoreDeleteSession()
*	iscsiCoreChangeSession()
*
* %I% %W% %D% %T%
*
*/

/*
* Change History
*
*    Date     Version Defect Initials Comments
* 2001/05/17  1.20    2337   KZM      After Reject Login, do not make new connection until after Reset.
*/

#include "iscsi_interfaces.h"
#include "iscsi_core_messages.h"
#include "iscsi_local_trace.h"
#include "iscsi_core_prot.h"

static u16 isid = 0xabcd;


iscsiLock_t	sessionsLock;
sessionInfo_t	*sessions[ISCSI_MAX_TARGETS];
u32	maxSessions = ISCSI_MAX_TARGETS;


void freeSessionResources(sessionInfo_t *si);
void destroyConnections(sessionInfo_t *si);


/*
* NAME: initSessions
*
* PURPOSE: initialize global variables used by session routines
*
* PARAMETERS: none
*
* RETURNS: none
*
* CALLED BY: iscsiCoreInitData()
*
* NOTES:
*/
void initSessions(void)
{
	iscsiOSlockInit(&sessionsLock);
	memset(sessions, 0, sizeof(sessions));
}


/*
* NAME: cleanupSessions
*
* PURPOSE: free up all sesseion resources
*
* PARAMETERS: none
*
* RETURNS: none
*
* CALLED BY: iscsiCoreCleanupData()
*
* NOTES:
*	During iscsi module unload, sessions should have been cleaned up already;
*	so we should have only global variables left to clean up.
*	Check, just to make sure.
*/
void cleanupSessions(void)
{
	u32 i;
	TRACE_ENTER;
	/* verify that all sessions have indeed exited */
	for (i = 0; i < maxSessions; i++) {
		if (NULL != sessions[i]) {
			iscsiCoreDeleteSession(i);
		}
	}
	iscsiOSlockDestroy(&sessionsLock);
	TRACE_EXIT;
}


/*
* NAME: getSessionStruct
*
* PURPOSE: map handle into pointer to a sessionInfo structure
*
* PARAMETERS:
*	sessionHandle - index of session; returned by iscsiCorecreateSession().
*
* RETURNS: pointer to corresponding sessionInfo structure; NULL if not defined.
*
* CALLED BY: various routines
*
* NOTES:
*/
sessionInfo_t * getSessionStruct(u32 sessionHandle)
{
	if (sessionHandle >= maxSessions) {
		return NULL;
	}
	return sessions[sessionHandle];
}



/*
* NAME: makeTcpConnections
*
* PURPOSE: establish TCP connections for specified session
*
* PARAMETERS:
*	si - pointer to sessionInfo structure containing connection information
*
* RETURNS: number of connections that exist for specified session
*
* CALLED BY: iscsiCoreCreateSession(), cmdQHandler()
*
* NOTES:
*	Perform Login iSCSI negotiation using session Login information stored in sessionInfo
*	structure.
*
* FUTURE ENHANCEMENTS:
*	Presently, each connection for a session is to the same address/port.
*	We should make this more general to allow different adddress/port for each connection.
*	We should then test which connections already exist, and try to create the
*	remaining connections.
*	Also, we should return more specific error conditions to inform
*	the calling function exactly why we were unable to make the
*	connections.
*	(ex. connect() failed, login rejected, malloc() failed, etc)
*/
u32 makeTcpConnections(sessionInfo_t *si)
{
	iscsiHeader_t message;
	iscsiHeader_t response;
	iscsiIrql_t irql;
	connInfo_t *ci;
	s32 rc = 0;
	u32 i;
	u32 length = 0;
	void *ptr;

	TRACE_ENTER;

	/* sanity check */
	PRINT(si->numActiveConnections);
	if (si->numActiveConnections == si->numMaxConnections) {
		return si->numActiveConnections;
	}

	/* build an iscsi login header */
	memset(&message, 0, sizeof(iscsiHeader_t));
	memset(&response, 0, sizeof(iscsiHeader_t));
	message.buf[OP_CODE_BYTE] = ISCSI_LOGIN_COMMAND;
	* (u32 *) &message.buf[LENGTH_BYTE] = htonl(si->loginParamsLen);
	* (u16 *) &message.buf[ISID_BYTE] = htons(si->isid);
	PRINT(si->loginParamsLen);
	PRBUF(&message.buf[LENGTH_BYTE], 4);

	PRINT(si->numActiveConnections);
	PRINT(si->numMaxConnections);
	for (i = 0; i < si->numMaxConnections; i++) {
		ci = &si->connections[i];
		PRHEX(ci);
		PRINT(ci->state);
		if ((CONNECTION_CONNECTED == ci->state) || (CONNECTION_LOGIN_REJECTED == ci->state)) {
			continue;
		}
		if (!si->numActiveConnections) {
			/* first connection */
			* (u32 *) &message.buf[CMD_RN_BYTE] = htonl(si->cmdRN); /* InitCmdRN */
			* (u16 *) &message.buf[TSID_BYTE] = 0; /* first connection */
		}
		else {
			* (u32 *) &message.buf[CMD_RN_BYTE] = 0;
			* (u16 *) &message.buf[TSID_BYTE] = htons(si->tsid);
		}
		PRINT(si->isid);
		PRINT(si->tsid);
		PRINT(* (u16 *) &message.buf[TSID_BYTE]);
		ci->sInfo = si;
		ci->state = CONNECTION_INITIALIZING;

		/* open a socket connection using an OS dependent call*/
		rc = iscsiOSmakeConnection(ci->addr, ci->portNum, ci->sock);
		PRINT(rc);
		if (rc < 0) {
			ci->state = CONNECTION_CONNECT_FAILED;
			continue;
		}

		/* send iscsi login header to server */
		rc = iscsiOSwriteToConnection(ci->sock, message.buf, ISCSI_HEADER_LEN, 0, 0);
		if (ISCSI_HEADER_LEN != rc) {
			ci->state = CONNECTION_ERROR;
			continue;
		}
		/* send login parameters */
		rc = iscsiOSwriteToConnection(ci->sock, si->loginParams, si->loginParamsLen, 0, 0);
		if (rc != (s32) si->loginParamsLen) {
			ci->state = CONNECTION_ERROR;
			continue;
		}

		/* wait for login response */
		PRMSG("waiting for login response");

		/* mpl we need to change the prototype for iscsiOSreadFromConnection so that we can */
		/* specify an optional timeout parameter.  This is necessary here, because we       */
		/* can't wait for a login response forever.  If the login response doesn't come     */
		/* in a reasonable amount of time we need to kill the login attempt and move on.    */


		/* iscsiOSreadFromConnection2 times out after 2 seconds, and returns -1 if a timeout occurs */
		rc = iscsiOSreadFromConnection2(ci->sock, response.buf, ISCSI_HEADER_LEN, 0 ,0);


		PRMSG("after login response");
		PRINT(rc);
		if (ISCSI_HEADER_LEN != rc) {
			ci->state = CONNECTION_ERROR;
			continue;
		}
		PRINT(response.buf[STATUS_BYTE]);
		
		/*
		Eat up any data beyond response header.
		In future, may have exchange of authentication info.
		*/
		length = ntohl(*(u32 *) &response.buf[LENGTH_BYTE]);
		PRINT(length);
		if (length > 0) {
			ptr = iscsiOSmalloc(length);
			if (NULL == ptr) {
				iscsiOSbreakConnection(ci->sock);
				ci->state = CONNECTION_NOT_CONNECTED;
				/* no use in going on to other connections */
				return si->numActiveConnections;
			}
			rc = iscsiOSreadFromConnection(ci->sock, ptr, length, 0, 0);
			iscsiOSfree(ptr);
			if (rc != (s32) length) {
				iscsiOSbreakConnection(ci->sock);
				ci->state = CONNECTION_ERROR;
				continue;
			}
		}

		/* verify that message type is Login Response; in future may also be Text response */
		PRHEX(response.buf[OP_CODE_BYTE]);
		if (response.buf[OP_CODE_BYTE] != ISCSI_LOGIN_RESPONSE) {
			PRMSG("bad Login Response");
			iscsiOSbreakConnection(ci->sock);
			ci->state = CONNECTION_LOGIN_ERROR;
			continue;
		}
		if (response.buf[STATUS_BYTE] != ACCEPT_LOGIN) {
			iscsiOSbreakConnection(ci->sock);
			ci->state = CONNECTION_LOGIN_REJECTED;
			continue;
		}

		if (!si->numActiveConnections) {
			si->tsid = ntohs(*(u16 *) &response.buf[TSID_BYTE]);
		}
		PRINT(si->tsid);
		si->expStatRN = ntohl(*(u32 *) &response.buf[STAT_RN_BYTE]);
		PRINT(si->expStatRN);
		/* xxx perform statRN processing */

		/* xxx check that expCmdRN matches */
		si->maxCmdRN = ntohl(*(u32 *) &response.buf[MAX_CMD_RN_BYTE]);
		/* xxx this isn't quite right */
		PRINT(si->maxCmdRN);

		/* create a thread to read input on this connection */
		PRMSG("before createThread");
		ci->readThreadId = iscsiOScreateThread(readConnectionHandler, ci);
		PRMSG("after createThread");
		PRINT(ci->readThreadId);
		if (ci->readThreadId <= 0) {
			iscsiOSbreakConnection(ci->sock);
			ci->readThreadId = 0;
			ci->state = CONNECTION_NOT_CONNECTED;
			continue;
		}

		ci->state = CONNECTION_CONNECTED;

		PRINT(si->numActiveConnections);
		iscsiOSlock(&si->lock, &irql);
		si->activeConnections[si->numActiveConnections] = ci;
		si->numActiveConnections++;
		iscsiOSunlock(&si->lock, irql);
		PRINT(si->numActiveConnections);
	}
	PRINT(si->numActiveConnections);

	TRACE_EXIT;
	return i;
}



/*
* NAME: breakTcpConnectionsAndEmptyQs
*
* PURPOSE: break connections of session and free all queued items
*
* PARAMETERS:
*	si - pointer to session structure
*
* RETURNS: none
*
* CALLED BY: freeSessionResources(), handleTaskesponse(), iscsiCoreResetBus(), iscsiCoreResetTarget()
*
* NOTES:
* Set SESSION_RESETTING flag while we are in the process of breaking
* TCP connections. The flag will be cleared when the last connection
* has been completely cleaned up.
* Then free up all commands in command Q,
* indicating that there is no longer a connection to the target.
* If no connections exist, we clear the SESSION_RESETTING
* flag when we are done.
* Break all TCP connections and force all read/receive threads to exit.
* For connections that were rejected for Login, allow to try to Login
* again.
*/
void breakTcpConnectionsAndEmptyQs(sessionInfo_t *si)
{
	s32	i;
	iscsiIrql_t irql;
	connInfo_t *ci;
	u32 connectionsExist = 0;

	TRACE_ENTER;

	PRHEX(si);
	PRINT(si->numActiveConnections);

	iscsiOSlock(&si->lock, &irql);
	si->flags |= SESSION_RESETTING;
	iscsiOSunlock(&si->lock, irql);

	/* empty out all of the queues */
	emptySessionQs(si);

	/* force all read threads to exit */
	/* don't hold spin lock when calling OS dependent call */
	PRINT(si->numActiveConnections);
	if (si->numActiveConnections) {
		connectionsExist = 1;
	}
	for (i = 0; i < (s32) si->numMaxConnections; i++) {
		ci = &si->connections[i];
		if ((CONNECTION_CONNECTED == ci->state) || (CONNECTION_INITIALIZING == ci->state)) {
			iscsiOSbreakConnection(ci->sock);
		}
		else if (CONNECTION_LOGIN_REJECTED == ci->state) {
			ci->state = CONNECTION_NOT_CONNECTED;
		}
	}

	if (!connectionsExist) {
		/* clear the flag that we are resetting */
		iscsiOSlock(&si->lock, &irql);
		si->flags &= ~SESSION_RESETTING;
		iscsiOSunlock(&si->lock, irql);
	}

	TRACE_EXIT;
}


/*
* NAME: iscsiCoreCreateSession
*
* PURPOSE: create an iSCSI session.
*
* PARAMETERS:
*	addr - address of iSCSI target
*	portNum - port number to use for the iSCSI connections
*	nConnections - number of connections to establish for this session.
*	loginParams - login parameters for session
*	loginParamsLen - length (number of bytes) of loginParams
*
* RETURNS: upon success - index (handle) of session; -1 upon error. 
*
* CALLED BY: OS dependent base driver routine
*
* NOTES:
* Verify that the specified session does not already exist.
* If it already exists, return existing session index.
* xxx
* The code is written to have the target defined even if
* no connecitons could be established.
* On each command to the target, we try to establish
* a TCP connection, if none had yet existed.
*
* FUTURE ENHANCEMENTS:
*	This should be changed to allow a different adddress/portNum for each connection.
*	We should also allow dynamic changing of connection information.
*/
s32 iscsiCoreCreateSession(u32 addr, u16 portNum, u32 nConnections, uchar *loginParams, u32 loginParamsLen)
{
	sessionInfo_t *si = NULL;
	u32 rc = 0;
	u32 size;
	u32 i;
	iscsiIrql_t irql;
	s32 index = -1;

	TRACE_ENTER;

	PRHEX(addr);
	PRINT(portNum);
	PRINT(nConnections);

	/* verify that session is not already in array of active sessions */
	iscsiOSlock(&sessionsLock, &irql);
	for (i = 0; i < maxSessions; i++) {
		if ((NULL != sessions[i]) &&
		  (sessions[i]->connections[0].addr == addr) &&
		  (sessions[i]->loginParamsLen == loginParamsLen) &&
		  SAME_STRING(sessions[i]->loginParams, loginParams, loginParamsLen)) {
			iscsiOSunlock(&sessionsLock, irql);
			return i;
		}
	}
	iscsiOSunlock(&sessionsLock, irql);

	/* allocate a sessionInfo structure for the new session */
	si = iscsiOSmalloc(sizeof(sessionInfo_t));
	PRHEX(si);
	if (NULL == si) {
		return -1;
	}
	memset(si, 0, sizeof(sessionInfo_t));

	/* allocate space to save connection information */
	size = sizeof(connInfo_t) * nConnections;
	PRINT(size);
	si->connections = iscsiOSmalloc(size);
	PRHEX(si->connections);
	if (NULL == si->connections) {
		iscsiOSfree(si);
		return -1;
	}
	memset(si->connections, 0, size);

	/* set address information for each connection */
	for (i = 0; i < nConnections; i++) {
		si->connections[i].addr = addr;
		si->connections[i].portNum = portNum;
		si->connections[i].state = CONNECTION_NOT_CONNECTED;
		rc = iscsiOSallocateConnection(addr, portNum, &si->connections[i].sock);
		if (rc < 0) {
			/* xxx what do we want to do? */
			si->connections[i].state = CONNECTION_BAD;
		}
	}

	size = sizeof(connInfo_t*) * nConnections;
	si->activeConnections = iscsiOSmalloc(size);
	PRHEX(si->activeConnections);
	if (NULL == si->activeConnections) {
		destroyConnections(si);
		iscsiOSfree(si->connections);
		iscsiOSfree(si);
		return -1;
	}
	memset(si->activeConnections, 0, size);

	si->stats = iscsiOSmalloc(sizeof(sessionStats_t));
	if (NULL == si->stats) {
		destroyConnections(si);
		iscsiOSfree(si->activeConnections);
		iscsiOSfree(si->connections);
		iscsiOSfree(si);
		return -1;
	}
	memset(si->stats, 0, sizeof(sessionStats_t));

	si->loginParamsLen = loginParamsLen;
	si->loginParams = iscsiOSmalloc(si->loginParamsLen);
	if (NULL == si->loginParams) {
		destroyConnections(si);
		iscsiOSfree(si->stats);
		iscsiOSfree(si->activeConnections);
		iscsiOSfree(si->connections);
		iscsiOSfree(si);
		return -1;
	}
	memcpy(si->loginParams, loginParams, si->loginParamsLen);
	PRINT(si->loginParamsLen);
	PRBUF(si->loginParams, si->loginParamsLen);

	/* initialize remaining fields of the session structure */
	si->numMaxConnections = nConnections;
	si->cmdRN = 1;
	si->maxCmdRN = 1;
	si->isid = isid++;
	iscsiOSlockInit(&si->lock);
	iscsiOSlockInit(&si->statsLock);
	iscsiOSeventInit(&si->cmdQEvent);
	iscsiOSeventInit(&si->exitEvent);
	Qinit(&si->cmdQ);
	Qinit(&si->taskManQ);
	Qinit(&si->rttQ);
	initPendingCmdTable(si);

	/* scan loginParams, and change ';' to '\0' to separate between tokens */
	/* Note: this disallows ';' from being part of a token value */
	for (i = 0; i < loginParamsLen; i++) {
		if (';' == si->loginParams[i]) {
			si->loginParams[i] = '\0';
		}
	}

	PRINT(si->cmdRN);
	PRINT(si->isid);

	/* add to array of active sessions */
	iscsiOSlock(&sessionsLock, &irql);
	/* find first empty location */
	for (i = 0; i < maxSessions; i++) {
		if (NULL == sessions[i]) {
			index = (u32) i;
			sessions[index] = si;
			break;
		}
	}
	iscsiOSunlock(&sessionsLock, irql);

	if (-1 == index) {
		/* no room for new session */
		/* free up the session */
		si->pendingCmdThreadId = 0;
		freeSessionResources(si);
		return -1;
	}

	/*
	We try to make connections here.
	If we are not successful, then a warning is printed in config.c
	that we have no connections.
	*/
	makeTcpConnections(si);

	PRMSG("before createThread");
	/* create thread to handle Qs of session */
	si->pendingCmdThreadId = iscsiOScreateThread(cmdQHandler, si);
	if (si->pendingCmdThreadId <= 0) {
		/* free up session resources and return -1 */
		si->pendingCmdThreadId = 0;
		sessions[index] = NULL;
		freeSessionResources(si);
		return -1;
	}

	TRACE_EXIT;
	return index;
}




/*
* NAME: emptySessionQs
*
* PURPOSE: free up all Q entries for specified session
*
* PARAMETERS:
*	si - pointer to sessionInfo structure of relevant session
*
* CALLEB BY: freeSessionResources()
*
* RETURNS: none
*
* NOTES:
*/
void emptySessionQs(sessionInfo_t *si)
{
	rttInfo_t	*rtt;
	pendingCmd_t	*cmd;
	taskManageInfo_t	*tm;

	TRACE_ENTER;
	PRHEX(si);

	/* clean out task management Q */
	while (si->taskManQ.count) {
		Qget(&si->taskManQ, tm, taskManageInfo_t);
		if (NULL == tm) {
			/* xxx this should never happen. log an error */
			break;
		}
		/* complete the operation with an error indication */
		PRHEX(tm);
		switch (tm->operation) {
			case ABORT_TASK:
				tm->result = ISCSI_ABORT_ERROR;
				break;
			default:
				/* xxx update this when we support operations besides RESET */ 
				tm->result = ISCSI_RESET_ERROR;
				break;
		}
		if (tm->flags & ISCSI_WAIT_FOR_COMPLETION) {
			/* wakeup thread that is waiting */
			PRMSG("waking up the task management requester");
			iscsiOSwakeup(&tm->event);
			/* awakened thread will release tm structure */
		}
		else {
			PRMSG("before iscsiOScompleteTaskMan");
			iscsiOScompleteTaskMan(tm->operation, tm->newScb, tm->result);
			Qput(&freeTaskManQ, tm, taskManageInfo_t);
		}
	}

	/* clean out rtt Q */
	while (si->rttQ.count) {
		Qget(&si->rttQ, rtt, rttInfo_t);
		if (NULL == rtt) {
			/* xxx should never happen. log an error */
			break;
		}
		Qput(&freeMsgHdrQ, rtt->rttHeader, iscsiHeader_t);
		Qput(&freeRttQ, rtt, rttInfo_t);
	}

	/* clean out cmd Q; return each command with an error */
	while (si->cmdQ.count) {
		Qget(&si->cmdQ, cmd, pendingCmd_t);
		if (NULL == cmd) {
			/* xxx should never happen. log an error */
			break;
		}
		iscsiOScompleteCommandReset(cmd->scb);
		ISCSI_INCREMENT_COUNTER(si, nCmdsReset);
		Qput(&freeCmdQ, cmd, pendingCmd_t);
	}

	TRACE_EXIT;
}


/*
* NAME: destroyConnections
*
* PURPOSE: free up all connection resources for a session
*
* PARAMETERS:
*	si - pointer to sessionInfo structure of relevant session
*
* RETURNS: none
*
* CALLED BY: freeSessionResource, iscsiCoreCreateSession
*
* NOTES:
*/
void destroyConnections(sessionInfo_t *si)
{
	u32 i;

	TRACE_ENTER;
	for (i = 0; i < si->numMaxConnections; i++) {
		if ((si->connections[i].sock != NULL) && (si->connections[i].state != CONNECTION_BAD)) {
			iscsiOSdestroyConnection(si->connections[i].sock);
		}
	}
	TRACE_EXIT;
}


/*
* NAME: freeSessionResources
*
* PURPOSE: free up all session resources, including threads, Qs, data structures.
*
* PARAMETERS:
*	si - pointer to sessionInfo structure of relevant session
*
* RETURNS: none
*
* CALLED BY: iscsiCoreDeleteSession()
*
* NOTES:
*	Cause all session threads to exit.
*	Then clean up the session structure.
*/
void freeSessionResources(sessionInfo_t *si)
{
	iscsiIrql_t irql;

	TRACE_ENTER;
	PRHEX(si);
	PRINT(si->pendingCmdThreadId);

	/* stop outgoing traffic by killing pending command Q thread */
	iscsiOSlock(&si->lock, &irql);
	si->flags |= SESSION_MUST_EXIT;
	iscsiOSwakeup(&si->cmdQEvent);

	/*
	Wait until command Q thread has exited.
	Sleep on special event which will be signalled by command Q
	thread when it exits.
	We can check that it has exited when its threadId field is
	set to 0.
	For SMPs, we should sleep on the event and give up the lock on
	the sessionInfo_t structure atomically, so as to avoid a race
	condition of the command Q thread exiting before we sleep here
	on the event.
	*/
	while (si->pendingCmdThreadId) {
		/* unlock and sleep atomically */
		iscsiOSunlockAndSleepAndLock(&si->lock, &si->exitEvent, &irql);
	}
	iscsiOSunlock(&si->lock, irql);

	breakTcpConnectionsAndEmptyQs(si);

	/* wait for all receive threads to exit */
	iscsiOSlock(&si->lock, &irql);

	/*
	Wait until all read threads have exited.
	*/
	while (si->numActiveConnections) {
		/* unlock and sleep atomically */
		iscsiOSunlockAndSleepAndLock(&si->lock, &si->exitEvent, &irql);
	}
	iscsiOSunlock(&si->lock, irql);
	PRINT(si->numActiveConnections);

	destroyConnections(si);

	/* clean up all of the session's data structures */
	Qfree(&si->cmdQ);
	Qfree(&si->taskManQ);
	Qfree(&si->rttQ);
	iscsiOSfree(si->connections);
	iscsiOSfree(si->activeConnections);
	iscsiOSfree(si->loginParams);
	iscsiOSfree(si->stats);
	iscsiOSlockDestroy(&si->lock);
	iscsiOSlockDestroy(&si->statsLock);
	iscsiOSeventDestroy(&si->cmdQEvent);
	iscsiOSeventDestroy(&si->exitEvent);
	iscsiOSfree(si);

	TRACE_EXIT;
}



/*
* NAME: iscsiCoreDeleteSession
*
* PURPOSE: called by OS dependent code to free up session resources.
*
* PARAMETERS:
*	sessionHandle - handle to sessionInfo structure of session to be deleted.
*
* RETURNS: none
*
* CALLED BY: OS dependent base driver routine, cleanupSessions()
*
* NOTES:
*	Verifies validity of handle and calls freeSessionResources() to free the resources.
*/
s32 iscsiCoreDeleteSession(u32 sessionHandle)
{
	sessionInfo_t *si;

	TRACE_ENTER;
	PRINT(sessionHandle);
	PRINT(maxSessions);

	/* remove session from list of valid sessions */
	if (sessionHandle >= maxSessions) {
		return -1;
	}
	PRHEX(sessions[sessionHandle]);
	si = sessions[sessionHandle];
	PRHEX(si);
	if (NULL == si) {
		return -1;
	}
	sessions[sessionHandle] = NULL;
	freeSessionResources(si);
	TRACE_EXIT;
	return 0;
}


/*
* NAME: iscsiCoreResetSessions
*
* PURPOSE: break all TCP connections of all sessions.
*
* PARAMETERS: none
*
* RETURNS: none
*
* CALLED BY: iscsi_ioctl
*
* NOTES:
*	TCP connection will be re-established when trying to send a
*	command to a particular target.
*/
void iscsiCoreResetSessions(void)
{
	s32 j;
	sessionInfo_t *si;

	TRACE_ENTER;
	for (j = 0; j < ISCSI_MAX_TARGETS; j++) {
		if (NULL != sessions[j]) {
			si = sessions[j];
			PRHEX(si);
			PRINT(si->numActiveConnections);
			breakTcpConnectionsAndEmptyQs(si);
		}
	}
	TRACE_EXIT;
}


/*
* NAME: iscsiCoreChangeSession
*
* PURPOSE: change attributes of an existing iSCSI session and reconnect to target
*
* PARAMETERS:
*	index - index (handle) of the existing iSCSI session
*	loginParams - login parameters for session
*	loginParamsLen - length (number of bytes) of loginParams
*
* RETURNS:  0 on success 
*          -1 on failure (most ccommonly this means there are active connections)
*
* CALLED BY: OS dependent base driver routine
*
* NOTES:
* Changes are allowed only if there are no active connections. 
* Currently only the login parameters can be changed.
*
* If called without new attributes, or if the new attributes are identical
* to the existing ones, the function just reconnects to the target. 
*/
s32 iscsiCoreChangeSession(u32 index, uchar *loginParams, u32 loginParamsLen)
{
	u32 i;
	iscsiIrql_t irql;
	uchar *newLoginParams = NULL;
	sessionInfo_t * si;
	connInfo_t *ci;

	TRACE_ENTER;

	/* verify that session exists */
	if (index >= maxSessions)
		return -1;
	if (NULL == sessions[index]) 
		return -1;

	si = sessions[index];

	if ((loginParams != NULL) && 
		(loginParamsLen > 0) &&
		((si->loginParamsLen != loginParamsLen) ||
		 !SAME_STRING(si->loginParams, loginParams, loginParamsLen)))
	{
		/* prepare a buffer of new login parameters */
		newLoginParams = iscsiOSmalloc(loginParamsLen);
		if (newLoginParams == NULL)
			return -1;
		memcpy(newLoginParams, loginParams, loginParamsLen);
		PRINT(loginParamsLen);
		PRBUF(newLoginParams, loginParamsLen);
		/* scan loginParams, and change ';' to '\0' to separate between tokens */
		/* Note: this disallows ';' from being part of a token value */
		for (i = 0; i < loginParamsLen; i++)
		{
			if (';' == newLoginParams[i])
				newLoginParams[i] = '\0';
		}

		/* switch the attributes in the session structure */
		iscsiOSlock(&si->lock, &irql);
		/* verify that there are no active connections */
		if (!si->numActiveConnections) 
		{
			/* first get rid of the old attribue buffer */
			if (si->loginParams != NULL)
				iscsiOSfree(si->loginParams);
			si->loginParamsLen = loginParamsLen;
			si->loginParams = newLoginParams;

			/* fix flag in connection structure(s) about Login Reject */
			for (i = 0; i < si->numMaxConnections; i++) {
				ci = &si->connections[i];
				if (CONNECTION_LOGIN_REJECTED == ci->state) {
					ci->state = CONNECTION_NOT_CONNECTED;
				}
			}
		}
		else  
		{
			/* active connections exist, do not make changes */
			iscsiOSunlock(&si->lock, irql);
			iscsiOSfree(newLoginParams);
			return -1;		
		}  
		iscsiOSunlock(&si->lock, irql);	
	}

	/* try to establish connections */
	makeTcpConnections(si);

	TRACE_EXIT;

	return 0;
}

