
/*
*
* iscsi initiator driver
*
* Authors: Kalman Meth, IBM Haifa Research Laboratory <meth@il.ibm.com>
*          Liran Schour, IBM Haifa Research Laboratory <lirans@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* socket.c
*
* iscsi socket routines - OS dependent.
*
* %I% %W% %D% %T%
*/

#include <asm/types.h>
#include <linux/sched.h>
#include <asm/processor.h>
#include <asm/errno.h>
#include <asm/uaccess.h>
#include <linux/socket.h>
#include <linux/uio.h>
#include <linux/blkdev.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/smp_lock.h>
#include "iscsi_local_types.h"
#include "iscsi_interfaces.h"
#include "iscsi_local_trace.h"
#include "iscsi_client.h"

extern u32		iscsiDriverExit;


/*
* NAME: iscsiOSreadFromConnection
*
* PURPOSE: read specified amount of data from specified connection
*
* PARAMETERS:
*	isock - pointer to socket structure of connection
*	buffer - buffer into which to read the data
*	len - number of bytes to read into buffer
*
* RETURNS:
*	number of bytes read
*	-1 for error
*
* CALLED BY: many functions
*
* NOTES:
*
* FUTURE ENHANCEMENTS:
*	We should probably add a flag to indicate a non-blocking read.
*       We should consider the offset argument in future
*/
s32 iscsiOSreadFromConnection(iscsiSock_t *isock, void *buffer, u32 len, u32 use_sg,u32 offset)
{
	struct iovec iov[ISCSI_MAX_SG];
	struct msghdr mhdr;
	struct scatterlist *pScatterList;
	mm_segment_t oldfs;
	int rc = 0;
	int sgCount;
	
	TRACE_ENTER;
	PRINT(len);
	PRHEX(buffer);

	/* sanity check */
	if (NULL == isock) {
		return -1;
	}
	if (NULL == buffer) {
		return -1;
	}

	// down(&isock->sem);
	if (isock->status != SOCK_STATUS_CONNECTED) {
		// up(&isock->sem);
	        iscsiOSbreakConnection(isock); 
	        return -1;
	}

	/* need to obtain the kernel lock for SMP */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )	
	lock_kernel();
#endif

	oldfs = get_fs();
	set_fs(get_ds()); /* buf is in kernel space */

	if(use_sg) {
	  sgCount = 0;
	  pScatterList = (struct scatterlist *)buffer;
	  while( sgCount < use_sg) {
	    iov[sgCount].iov_base = pScatterList->address;
	    iov[sgCount].iov_len = pScatterList->length;
	    sgCount += 1;
	    pScatterList += 1;
	  }
	}
	else {
	  sgCount = 1;
	  iov[0].iov_base = buffer;
	  iov[0].iov_len = len;
	}
	
	memset(&mhdr, 0, sizeof(struct msghdr));
	mhdr.msg_iov = iov;
	mhdr.msg_iovlen = sgCount;

	
	rc = sock_recvmsg(isock->sock, &mhdr, len, MSG_WAITALL);

	set_fs(oldfs);

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
	unlock_kernel();
#endif
	// up(&isock->sem);
	
	// break TCP connection in any case of error
	if(rc < 0) {
	  iscsiOSbreakConnection(isock);
	}

	TRACE_EXIT;
	return rc;
}


void iscsi_read_timeout(unsigned long task)
{
  
  TRACE_ENTER;

  send_sig(SIGKILL, (struct task_struct *)task, 1);

  TRACE_EXIT;

}

/*
* NAME: iscsiOSreadFromConnection2
*
* PURPOSE: read specified amount of data from specified connection
*         with timeout for waiting for answer.
*
* PARAMETERS:
*	isock - pointer to socket structure of connection
*	buffer - buffer into which to read the data
*	len - number of bytes to read into buffer
*
* RETURNS:
*	number of bytes read
*	-1 for error
*
* CALLED BY: many functions
*
* NOTES:
*
* FUTURE ENHANCEMENTS:
*       We should merge iscsiOSwriteToConnection with this function by adding 
*       a flag indicating if we want to activate a timer.
*/
s32 iscsiOSreadFromConnection2(iscsiSock_t *isock, void *buffer, u32 len, u32 use_sg,u32 offset)
{
	struct iovec iov[ISCSI_MAX_SG];
	struct msghdr mhdr;
	struct scatterlist *pScatterList;
	mm_segment_t oldfs;
	int rc = 0;
	int sgCount;
	struct timer_list   read_timeout;
	
	TRACE_ENTER;
	PRINT(len);
	PRHEX(buffer);

	/* sanity check */
	if (NULL == isock) {
		return -1;
	}
	if (NULL == buffer) {
		return -1;
	}

	// down(&isock->sem);
	if (isock->status != SOCK_STATUS_CONNECTED) {
		// up(&isock->sem);
		iscsiOSbreakConnection(isock);
	}

	// initialize timeout
	init_timer(&read_timeout);
	read_timeout.data = (unsigned long) current;
	read_timeout.expires = jiffies + ISCSI_READ_TIMEOUT;
	read_timeout.function = (void (*)(unsigned long))iscsi_read_timeout;
	
	add_timer(&read_timeout);
	
	/* need to obtain the kernel lock for SMP */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )	
	lock_kernel();
#endif

	oldfs = get_fs();
	set_fs(get_ds()); /* buf is in kernel space */

	if(use_sg) {
	  sgCount = 0;
	  pScatterList = (struct scatterlist *)buffer;
	  while( sgCount < use_sg) {
	    iov[sgCount].iov_base = pScatterList->address;
	    iov[sgCount].iov_len = pScatterList->length;
	    sgCount += 1;
	    pScatterList += 1;
	  }
	}
	else {
	  sgCount = 1;
	  iov[0].iov_base = buffer;
	  iov[0].iov_len = len;
	}
	
	memset(&mhdr, 0, sizeof(struct msghdr));
	mhdr.msg_iov = iov;
	mhdr.msg_iovlen = sgCount;

	
	rc = sock_recvmsg(isock->sock, &mhdr, len, MSG_WAITALL);

	set_fs(oldfs);

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
	unlock_kernel();
#endif
	// up(&isock->sem);

	del_timer(&read_timeout);
	
	// break TCP connection in any case of error
	if(rc < 0) {
	  iscsiOSbreakConnection(isock);
	}

	TRACE_EXIT;
	return rc;
}


/*
* NAME: iscsiOSwriteToConnection
*
* PURPOSE: write specified amount of data to specified connection
*
* PARAMETERS:
*	isock - pointer to socket structure of connection
*	buffer - buffer from which to write the data
*	len - number of bytes to write from buffer
*
* RETURNS:
*	number of bytes written
*	-1 for error
*
* CALLED BY: many functions
*
* NOTES:
*
* FUTURE ENHANCEMENTS:
*	We should probably add a flag to indicate a non-blocking write.
*	We should probably also add an interface to determine whether a
*	connection is available for writing.
*       We should consider the offset argument in future
*/
s32 iscsiOSwriteToConnection(iscsiSock_t *isock, void *buffer, u32 len, u32 use_sg, u32 offset)
{
	int rc = 0;
	struct iovec iov[ISCSI_MAX_SG];
	struct msghdr mhdr;
	struct scatterlist *pScatterList;
	int    sgCount;
	mm_segment_t oldfs;

	TRACE_ENTER;
	PRINT(len);
	PRHEX(buffer);

	/* sanity check */
	if (NULL == isock) {
		return -1;
	}

	// down(&isock->sem);
	if (isock->status != SOCK_STATUS_CONNECTED) {
		// up(&isock->sem);
		iscsiOSbreakConnection(isock);
	        return -1;
	}

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
	lock_kernel();
#endif
	current->flags |= PF_MEMALLOC;
	oldfs = get_fs();
	set_fs(get_ds()); /* buf is in kernel space */
	
	if(use_sg) {
	  sgCount = 0;
	  pScatterList = (struct scatterlist *)buffer;
	  while( sgCount < use_sg) {
	    iov[sgCount].iov_base = pScatterList->address;
	    iov[sgCount].iov_len = pScatterList->length;
	    sgCount += 1;
	    pScatterList += 1;
	  }
	}
	else {
	  sgCount = 1;
	  iov[0].iov_base = buffer;
	  iov[0].iov_len = len;
	}

	memset(&mhdr, 0, sizeof(struct msghdr));
	mhdr.msg_iov = iov;
	mhdr.msg_iovlen = sgCount;	
	
	rc = sock_sendmsg(isock->sock, &mhdr, len);
	
	set_fs(oldfs);
	current->flags &= ~PF_MEMALLOC;	
	
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
	unlock_kernel();
#endif

	// up(&isock->sem);

	// break TCP connection in any case of error
	if(rc < 0) {
	  iscsiOSbreakConnection(isock);
	}
	
	TRACE_EXIT;
	return rc;
}


/*
* NAME: iscsiOSallocateConnection
*
* PURPOSE: allocate a TCP socket connection structure
*
* PARAMETERS:
*	addr - peer address to which to connect (not used)
*	portNum - peer port number to use (not used)
*	isock (OUT) - pointer to resulting socket structure
*
* RETURNS:
*	0 - success
*	-1 - failure
*
* CALLED BY:
*
* NOTES:
*/
int iscsiOSallocateConnection(u32 addr, u16 portNum, iscsiSock_t **isock)
{
	TRACE_ENTER;
	(*isock) = iscsiOSmalloc(sizeof(iscsiSock_t));
	if (NULL == (*isock)) {
		TRACE_EXIT;
		return -1;
	}
	memset((*isock), 0, sizeof(iscsiSock_t));
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )	
	(*isock)->sem = MUTEX;
#else
	init_MUTEX(&(*isock)->sem);
#endif
	(*isock)->status = SOCK_STATUS_INVALID;
	iscsiOSeventInit(&(*isock)->event);

	TRACE_EXIT;
	return 0;
}

/*
* NAME: socketConnect
*
* PURPOSE: create TCP socket connection
*
* PARAMETERS:
*	isock - pointer to iscsi socket structure
*
* RETURNS:
*	0 - success
*	<0 - failure
*
* CALLED BY: iscsiOSmakeConnection(), connectionManager()
*
* NOTES: 
*       
*/
int socketConnect(iscsiSock_t *isock)
{
  int rc = 0;
  struct socket *sock;
  
  TRACE_ENTER;
  
  if(isock == NULL) {
    return -1;
  }
  
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
  lock_kernel();
#endif  

  if (NULL != isock->sock) {
       sock_release(isock->sock);
	    isock->sock = NULL;
  }

  /* open a socket connection */
  rc = sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
  PR2("rc = %d, sock = %x \n", rc, (int) sock);
  if (rc >= 0) {
    
    PRHEX(sock);
    
    /* set up socket option */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
    sock->sk->nonagle = 1;
#endif
    
    sock->sk->allocation = GFP_ATOMIC;
    
    
    /* try to connect */
    /*  last parameter of connect() is for flags (which may be O_NONBLOCK) */
    rc = (sock)->ops->connect(sock, (struct sockaddr *) &isock->sin, sizeof(isock->sin), 0);
    
    PR2("rc = %d, sock = %x \n", rc, (int) sock);
    
    if (rc < 0) {
      
      sock_release(sock);
      
    }
    else { // success
      
      isock->sock = sock;
      
    }  
  }
  
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )
      unlock_kernel();
#endif	  
  
  TRACE_EXIT;
  return rc;

}

/*
* NAME: connectionManager
*
* PURPOSE: protect a TCP connection
*
* PARAMETERS:
*	arg - pointer to iscsi socket structure
*
* RETURNS:
*	0 - success
*	-1 - failure
*
* CALLED BY: iscsiOSmakeConnection()
*
* NOTES: should be replaced by a handler per session that will insure 
*        all sessions connections (should be in Core code)
*/
int connectionManager (void *arg)
{
  iscsiSock_t *isock = (iscsiSock_t *)arg;
  int         rc;
  
  TRACE_ENTER;
  if(isock == NULL) {
    return -1;
  }

  down(&isock->sem);
  // don't release the driver until all threads are done
  isock->threadIsActive = 1;
  isock->threadId = current;
  if (SOCK_STATUS_CONNECTED != isock->status) {
    // we need to connect 
    isock->status = SOCK_STATUS_CONNECTING;

    up(&isock->sem);
      
    rc = socketConnect(isock);
    
    down(&isock->sem);
    if(SOCK_STATUS_CONNECTING == isock->status && rc >= 0) {
	isock->status = SOCK_STATUS_CONNECTED;
    }
    else {
      isock->status = SOCK_STATUS_UNCONNECTED;
    }
    up(&isock->sem);
    
    iscsiOSwakeup(&isock->event);
    
  }
  else {
    up(&isock->sem);
  }
  
  down(&isock->sem);
  isock->threadIsActive = 0;
  up(&isock->sem);

  TRACE_EXIT;
  return 0;
}


/*
* NAME: iscsiOSmakeConnection
*
* PURPOSE: make a TCP connection
*
* PARAMETERS:
*	addr - peer address to which to connect
*	portNum - peer port number to use
*	sock - pointer to socket structure
*
* RETURNS:
*	0 - success
*	-1 - failure
*
* CALLED BY: makeTcpConnections()
*
* NOTES:
*/
int iscsiOSmakeConnection(u32 addr, u16 portNum, iscsiSock_t *isock)
{
        int rc = 0;
	
	TRACE_ENTER;
	PRHEX(addr);
	PRINT(portNum);
	PRHEX(isock);

	/* check if existing sock is already valid */
	down(&isock->sem);
	if (!isock->threadIsActive) {

	  up(&isock->sem);
	
	  /* set up connection parameters */
	  memset(&isock->sin, 0, sizeof(isock->sin));
	  isock->sin.sin_family = AF_INET;
	  isock->sin.sin_addr.s_addr = addr;
	  isock->sin.sin_port = htons(portNum);	  
	  
	  //socketConnect(isock);
	  
	  iscsiOScreateThread(connectionManager, isock);
	  
	  // we want to wait here to give the manager a chance to connect
	  PRMSG("before sleep");
	  sleep_on_timeout(&isock->event, 3 * HZ);
	  PRMSG("wakeup");
	  
	  down(&isock->sem);
	}// SOCK_STATUS_UNCONNECTED
	

	if( SOCK_STATUS_CONNECTED == isock->status) { 
	  rc = 0; 
	}
	else {
	  rc = -1;
	}
	
	
	up(&isock->sem); 
	
	
	TRACE_EXIT;
	return rc;
}



#define SHUT_RDWR 2

/*
* NAME: iscsiOSbreakConnection
*
* PURPOSE: break a TCP connection
*
* PARAMETERS:
*	sock - pointer to socket structure of connection
*
* RETURNS: none
*
* CALLED BY: cleanupTcpConnections(), makeTcpConnections()
*
* NOTES:
* This routine simply shuts down the socket.
* It does not destroy the socket structure.
*/
void iscsiOSbreakConnection(iscsiSock_t *isock) {
	TRACE_ENTER;
	PRHEX(isock);

	down(&isock->sem);
	if (SOCK_STATUS_CONNECTED != isock->status) {
	  if(SOCK_STATUS_CONNECTING == isock->status) {
	    isock->status = SOCK_STATUS_UNCONNECTED;
	  }
	  up(&isock->sem);
	  return;
	}
	
	isock->status = SOCK_STATUS_UNCONNECTED;
	up(&isock->sem);
	
	PRMSG("before shutdown");

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )	
	lock_kernel();
#endif

	isock->sock->ops->shutdown(isock->sock, SHUT_RDWR);

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )	
	unlock_kernel();
#endif
	// up(&isock->sem);
	TRACE_EXIT;
}


/*
* NAME: iscsiOSdestroyConnection
*
* PURPOSE: free up resources used by socket structure
*
* PARAMETERS:
*	isock - pointer to socket structure of free
*
* RETURNS: none
*
* CALLED BY: readConnectionhandler(), makeTcpConnections()
*
* NOTES:
* This routine destroys the socket structure and frees its memory.
*/
void iscsiOSdestroyConnection(iscsiSock_t *isock) {
	TRACE_ENTER;
	PRHEX(isock);
	
	if (NULL != isock) {
	  // we want to wait until connecting thread is out
	  down(&isock->sem);
	  while(isock->threadIsActive) {
	    up(&isock->sem);
	    send_sig(SIGKILL, (struct task_struct *)isock->threadId, 1);
	    current->state = TASK_UNINTERRUPTIBLE;
	    schedule_timeout(3 * HZ);
	    down(&isock->sem);
	  }
	  up(&isock->sem);
	  
	  if(NULL != isock->sock) {
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )	
	  lock_kernel();
#endif  
	  sock_release(isock->sock);
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) )	
	  unlock_kernel();
#endif	  
	  }
	}
	iscsiOSfree(isock);
	isock = NULL;

	TRACE_EXIT;
	
}
