
/*
*
* iscsi initiator driver
*
* Author: Kalman Meth, IBM Haifa Research Laboratory <meth@il.ibm.com>
*
* Copyright (c) IBM Corporation 2001 
* Licensed Material - Program Property of IBM
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. Neither the name of IBM Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
* HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* task.c
*
* Routines to handle Task Mangement operations (abort, reset, etc).
*
* %I% %W% %D% %T%
*/

#include "iscsi_interfaces.h"
#include "iscsi_core_types.h"
#include "iscsi_core_messages.h"
#include "iscsi_local_trace.h"
#include "iscsi_core_queue.h"
#include "iscsi_core_prot.h"



/*
* NAME: removeCmdFromQ
*
* PURPOSE: Remove specified SCSI command from command Q.
*
* PARAMETERS:
*	cmdQ - Q from which to remove command, if found
*	scb - pointer to OS-specific SCSI command block
*
* RETURNS:
*	pointer to corresponding pending command structure
*	NULL if not found
*
* CALLED BY: iscsiCoreAbortCmd()
*
* NOTES:
*/
pendingCmd_t *removeCmdFromQ(Q_t *cmdQ, SCB_t *scb)
{
	u32 irql;
	pendingCmd_t *cmd;
	pendingCmd_t *prev = NULL;

	TRACE_ENTER;
	PRHEX(scb);

	if (NULL == cmdQ->head) {
		TRACE_EXIT;
		return NULL;
	}

	iscsiOSlock(&cmdQ->lock, &irql);
	for (cmd = cmdQ->head; NULL != cmd; prev = cmd, cmd = cmd->forw) {
		if (cmd->scb == scb) {
			break;
		}
	}
	if (NULL != cmd) {
		if (cmd == cmdQ->head) {
			/* item was at head of Q */
			cmdQ->head = cmd->forw;
		}
		else {
			prev->forw = cmd->forw;
		}

		if (cmd == cmdQ->tail) {
			cmdQ->tail = prev;
		}

		cmdQ->count--;
		cmd->forw = NULL;
	}
	iscsiOSunlock(&cmdQ->lock, irql);
	PRHEX(cmd);

	TRACE_EXIT;
	return cmd;
}


/*
* NAME: iscsiCoreAbortCmd
*
* PURPOSE: abort a previously queued or started SCSI command
*
* PARAMETERS:
*	oldScb - SCSI Command Block of command to abort
*	newScb - SCSI Command Block of abort request (in some systems)
*	sessionHandle - session to which aborted command belongs
*	flags - indicates whether we should wait for result
*
* RETURNS:
*	ISCSI_ABORT_SUCCESS - successfully aborted specified command
*	ISCSI_ABORT_PENDING - (asynchronous) abort operation in progress
*	ISCSI_ABORT_CMD_NOT_FOUND - specified command not found; probably already completed
*	ISCSI_ABORT_ERROR - session not defined or some other error
*
* CALLED BY: OS dependent routine
*
* NOTES:
* We first check if the command to abort is still in the command Q.
* If yes, we handle the abort request immediately.
* If no, check in the session's pending command table.
* If found in pending command table, forward abort request to target.
* If flag indicates to wait for completion of abort operation,
* sleep on event until response arrives from target as to completion
* of the abort operation.
*/
u32 iscsiCoreAbortCmd(SCB_t *oldScb, SCB_t *newScb, u32 sessionHandle, u32 flags)
{
	pendingCmd_t		*cmd;
	sessionInfo_t		*si;
	taskManageInfo_t	*tm;
	iscsiIrql_t		irql;

	TRACE_ENTER;

	si = getSessionStruct(sessionHandle);
	if (NULL == si) {
		return ISCSI_ABORT_ERROR;
	}
	ISCSI_INCREMENT_COUNTER(si, nAbortCmd);

	/*
	check if command is in Command queue;
	if so, complete the command and complete the abort operation.
	*/
	cmd = removeCmdFromQ(&si->cmdQ, oldScb);
	PRHEX(cmd);
	if (NULL != cmd) {
		/* cancel command that is to be aborted */
		iscsiOScompleteCommandAbort(oldScb);
		ISCSI_INCREMENT_COUNTER(si, nCmdsAborted);
		Qput(&freeCmdQ, cmd, pendingCmd_t);
		TRACE_EXIT;
		return ISCSI_ABORT_SUCCESS;
	}

	/*
	Command was not in command Q.
	Check the pending command table.
	*/
	cmd = findCmdInTable(si, oldScb);
	PRHEX(cmd);
	if (NULL == cmd) {
		/* command not found; nothing more we can do */
		ISCSI_INCREMENT_COUNTER(si, nCmdsNotFound);
		TRACE_EXIT;
		return ISCSI_ABORT_CMD_NOT_FOUND;
	}

	/*
	Note: xxx There is a race condition here.
	We will use the information pointed to by cmd to create
	a Task Management request to be sent to target.
	In the meantime, it is possible that the command will
	complete, and the structure pointed to by cmd will be freed
	and then used for another command. It might appear that
	this can cause us to corrupt our internal data structures.
	However, since the command is no longer on the target, the target
	will respond that the Abort operation failed, and we will thus
	not touch the new structure that is pointed to by cmd.
	*/

	/*
	If have no active connections, cancel the aborted command.
	*/
	if (!si->numActiveConnections) {
		/* remove command from Pending Command Table */
		/* is there a race condition here? */
		cmd = removeCmdFromTable(si, cmd);
		if (NULL != cmd) {
			iscsiOScompleteCommandAbort(cmd->scb);
			ISCSI_INCREMENT_COUNTER(si, nCmdsAborted);
			Qput(&freeMsgHdrQ, cmd->header, iscsiHeader_t);
			Qput(&freeCmdQ, cmd, pendingCmd_t);
			TRACE_EXIT;
			return ISCSI_ABORT_SUCCESS;
		}
		else {
			TRACE_EXIT;
			return ISCSI_ABORT_CMD_NOT_FOUND;
		}
	}

	/* place Task Management request on Q to be sent to target */
	Qget(&freeTaskManQ, tm, taskManageInfo_t);
	if (NULL == tm) {
		return ISCSI_ABORT_ERROR;
	}
	PRHEX(tm);
	tm->oldScb = oldScb;
	tm->newScb = newScb;
	tm->pendingCmd = cmd;
	tm->operation = ABORT_TASK;
	tm->flags = flags;
	tm->result = 0;
	memset(tm->lun, 0, sizeof(iscsiLUN_t));
	Qput(&si->taskManQ, tm, taskManageInfo_t);
	ISCSI_INCREMENT_COUNTER(si, nAbortsQueued);

	/* wakeup the command handler to send request to target */
	iscsiOSlock(&si->lock, &irql);
	iscsiOSwakeup(&si->cmdQEvent);

	/* check to see if we must wait for response */
	if (flags & ISCSI_WAIT_FOR_COMPLETION) {
		u32 result;

		/* sleep on event in tm structure */
		PRMSG("going to sleep");
		iscsiOSunlockAndSleepAndLock(&si->lock, &tm->event, &irql);
		iscsiOSunlock(&si->lock, irql);
		PRMSG("awakened");

		/* when wakeup, take result from tm structure */
		result = tm->result;

		if (!result) {
			result = ISCSI_ABORT_ERROR;
		}

		/* free the tm structure */
		Qput(&freeTaskManQ, tm, taskManageInfo_t);
		TRACE_EXIT;
		return result;
	}
	iscsiOSunlock(&si->lock, irql);
	TRACE_EXIT;
	return ISCSI_ABORT_PENDING;
}


/*
* NAME: iscsiCoreResetDevice
*
* PURPOSE: reset specified device.
*
* PARAMETERS:
*	sessionHandle - session to which reset device belongs
*	scb - SCSI Command Block of reset request (in some systems)
*	lun - logical unit number of device to reset
*	flags - indicates whether we should wait for result
*
* RETURNS:
*	ISCSI_RESET_SUCCESS - successfully reset specified device
*	ISCSI_RESET_PENDING - (asynchronous) reset operation in progress
*	ISCSI_RESET_ERROR - session not defined or some other error
*
* CALLED BY: OS dependent routine
*
* NOTES:
* Send a message to the target to reset the device.
* When the device has been reset,
* complete all outstanding requests for specified device with an
* indication that the request was cancelled due to a reset operation.
* This will be performed in the receive thread.
*/
u32 iscsiCoreResetDevice(u32 sessionHandle, SCB_t *scb, iscsiLUN_t lun, u32 flags)
{
	sessionInfo_t		*si;
	taskManageInfo_t	*tm;
	pendingCmd_t		*cmd;
	pendingCmd_t		*prev;
	iscsiIrql_t		irql;
	Q_t			*cmdQ;

	TRACE_ENTER;
	PRHEX(lun[1]);

	si = getSessionStruct(sessionHandle);
	if (NULL == si) {
		return ISCSI_RESET_ERROR;
	}
	PRHEX(si);
	ISCSI_INCREMENT_COUNTER(si, nResetDevice);

	/* xxx cancel all Abort operations that are pending for specified device */

	/* cancel all commands in command Q for specified device */
	cmdQ = &si->cmdQ;
restart:
	prev = NULL;
	iscsiOSlock(&cmdQ->lock, &irql);
	for (cmd = cmdQ->head; NULL != cmd; ) {
		if (SAME_LUN(cmd->lun, lun)) {
			/* remove item from Q */
			if (cmd == cmdQ->head) {
				/* we are at head */
				cmdQ->head = cmd->forw;
				if (NULL == cmdQ->head) {
					cmdQ->tail = NULL;
				}
			}
			else {
				prev->forw = cmd->forw;
			}
			cmdQ->count--;

			/* don't hold spin lock when calling OS dependent routines */
			iscsiOSunlock(&cmdQ->lock, irql);
			PRHEX(cmd);
			/* xxx do we have to differentiate between resetBus and resetDevice? */
			iscsiOScompleteCommandReset(cmd->scb);
			ISCSI_INCREMENT_COUNTER(si, nCmdsReset);
			Qput(&freeCmdQ, cmd, pendingCmd_t);

			/* restart the loop */
			goto restart;
		}
		else {
			prev = cmd;
			cmd = cmd->forw;
		}
	}
	iscsiOSunlock(&cmdQ->lock, irql);

	/*
	If have no active connections, nothing we can do
	*/
	if (!si->numActiveConnections) {
		TRACE_EXIT;
		return ISCSI_RESET_ERROR;
	}

	/* place Task Management request on Q to be sent to target */
	Qget(&freeTaskManQ, tm, taskManageInfo_t);
	if (NULL == tm) {
		TRACE_EXIT;
		return ISCSI_RESET_ERROR;
	}
	tm->oldScb = NULL;
	tm->newScb = scb;
	tm->pendingCmd = NULL;
	tm->operation = LOGICAL_UNIT_RESET;
	tm->flags = flags;
	tm->result = 0;
	memcpy(tm->lun, lun, sizeof(iscsiLUN_t));
	Qput(&si->taskManQ, tm, taskManageInfo_t);

	/* wakeup the command handler to send request to target */
	iscsiOSlock(&si->lock, &irql);
	iscsiOSwakeup(&si->cmdQEvent);

	/* check to see if we must wait for response */
	if (flags & ISCSI_WAIT_FOR_COMPLETION) {
		u32 result;

		/* sleep on event in tm structure */
		PRMSG("going to sleep");
		iscsiOSunlockAndSleepAndLock(&si->lock, &tm->event, &irql);
		iscsiOSunlock(&si->lock, irql);
		PRMSG("awakened");

		/* when wakeup, take result from tm structure */
		result = tm->result;

		if (!result) {
			result = ISCSI_RESET_ERROR;
		}

		/* free the tm structure */
		Qput(&freeTaskManQ, tm, taskManageInfo_t);
		TRACE_EXIT;
		return result;
	}
	iscsiOSunlock(&si->lock, irql);
	TRACE_EXIT;
	return ISCSI_RESET_PENDING;
}


/*
* NAME: iscsiCoreResetBus
*
* PURPOSE: reset interface to specified target
*
* PARAMETERS:
*	sessionHandle - session to be reset
*
* RETURNS:
*	ISCSI_RESET_SUCCESS - successfully reset specified session
*	ISCSI_RESET_ERROR - session not defined or some other error
*
* CALLED BY: OS dependent routine
*
* NOTES:
* Complete all outstanding requests for specified target with an
* indication that the request was cancelled due to a reset operation.
* Don't wait for interaction with target.
* Cancel everything locally, break TCP connections, and re-establish
* the TCP connections.
* It is assumed that no new commands will be started for the specified
* session until the reset operation is completed.
*/
u32 iscsiCoreResetBus(u32 sessionHandle, SCB_t *scb, u32 flags)
{
	sessionInfo_t		*si;
	iscsiIrql_t		irql;

	TRACE_ENTER;

	PRINT(sessionHandle);
	si = getSessionStruct(sessionHandle);
	PRHEX(si);
	if (NULL == si) {
		return ISCSI_RESET_ERROR;
	}
	ISCSI_INCREMENT_COUNTER(si, nResetBus);

	PRINT(si->numActiveConnections);
	if (flags == ISCSI_KEEP_CONNECTION_ALIVE)
	{
		iscsiOSlock(&si->lock, &irql);
		emptySessionQs(si);
		iscsiOSunlock(&si->lock, irql);
	}
	else
	{
		breakTcpConnectionsAndEmptyQs(si);
	}
	TRACE_EXIT;
	return ISCSI_RESET_SUCCESS;
}


/*
* NAME: iscsiCoreResetTarget
*
* PURPOSE: send message to target to reset itself
*
* PARAMETERS:
*	sessionHandle - session of target to be reset
*	scb - SCSI Command Block of reset request (in some systems)
*
* RETURNS:
*	ISCSI_RESET_SUCCESS - successfully reset specified session
*	ISCSI_RESET_PENDING - (asynchronous) reset operation in progress
*	ISCSI_RESET_ERROR - session not defined or some other error
*
* CALLED BY: OS dependent routine
*
* NOTES:
*/
u32 iscsiCoreResetTarget(u32 sessionHandle, SCB_t *scb, u32 flags)
{
	sessionInfo_t		*si;
	taskManageInfo_t	*tm;
	iscsiIrql_t		irql;

	TRACE_ENTER;

	si = getSessionStruct(sessionHandle);
	if (NULL == si) {
		TRACE_EXIT;
		return ISCSI_RESET_ERROR;
	}
	ISCSI_INCREMENT_COUNTER(si, nResetTarget);

	/*
	If have no active connections, simply clean up data structures.
	*/
	PRINT(si->numActiveConnections);
	if (!si->numActiveConnections) {
		breakTcpConnectionsAndEmptyQs(si);

		TRACE_EXIT;
		return ISCSI_RESET_SUCCESS;
	}

	/* place Task Management request on Q to be sent to target */
	Qget(&freeTaskManQ, tm, taskManageInfo_t);
	PRHEX(tm);
	if (NULL == tm) {
		TRACE_EXIT;
		return ISCSI_RESET_ERROR;
	}
	tm->oldScb = NULL;
	tm->newScb = scb;
	tm->pendingCmd = NULL;
	tm->operation = TARGET_RESET;
	tm->flags = flags;
	tm->result = 0;
	memset(tm->lun, 0, sizeof(iscsiLUN_t));
	Qput(&si->taskManQ, tm, taskManageInfo_t);

	/* wakeup the command handler to send request to target */
	iscsiOSlock(&si->lock, &irql);
	iscsiOSwakeup(&si->cmdQEvent);

	/* check to see if we must wait for response */
	if (flags & ISCSI_WAIT_FOR_COMPLETION) {
		u32 result;

		/* sleep on event in tm structure */
		PRMSG("going to sleep");
		iscsiOSunlockAndSleepAndLock(&si->lock, &tm->event, &irql);
		iscsiOSunlock(&si->lock, irql);
		PRMSG("awakened");

		/* when wakeup, take result from tm structure */
		result = tm->result;
		PRINT(result);

		if (!result) {
			result = ISCSI_RESET_ERROR;
		}

		/* free the tm structure */
		Qput(&freeTaskManQ, tm, taskManageInfo_t);
		TRACE_EXIT;
		return result;
	}
	iscsiOSunlock(&si->lock, irql);
	TRACE_EXIT;
	return ISCSI_RESET_PENDING;
}


/*
* NAME: finishTaskManOperation
*
* PURPOSE: complete common processing of a Task Management operation
*
* PARAMETERS:
*	tm - pointer to Task Mangement information structure
*
*
* RETURNS: none
*
* CALLED BY: many functions
*
* NOTES:
*/
void finishTaskManOperation(sessionInfo_t *si, taskManageInfo_t *tm)
{
	iscsiIrql_t irql;

	TRACE_ENTER;
	PRHEX(tm);
	PRINT(tm->flags);
	PRINT(tm->result);
	if (tm->flags & ISCSI_WAIT_FOR_COMPLETION) {
		/* wakeup thread that is waiting */
		PRMSG("waking up the task management requester");
		iscsiOSlock(&si->lock, &irql);
		iscsiOSwakeup(&tm->event);
		iscsiOSunlock(&si->lock, irql);
		/* awakened thread will release tm structure */
	}
	else {
		PRMSG("before iscsiOScompleteTaskMan");
		iscsiOScompleteTaskMan(tm->operation, tm->newScb, tm->result);
		Qput(&freeTaskManQ, tm, taskManageInfo_t);
	}
	TRACE_EXIT;;
	return;
}

